# Simple support for organizing vcs branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

# TODO: write tests for actions!

"""Actions to execute as project-wide commands."""

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
import os
import shlex
import tempfile

from bzrlib import (
    errors,
    osutils,
    urlutils,
    )
from bzrlib.commands import run_bzr_catch_user_errors

from bzrlib.plugins.scmproj import config, project
from bzrlib.plugins.scmproj.vcs import vcs_reg, bzr_vcs
""")


def url_for_display(url):
    """Return human-readable URL or local path for file:/// URLs.
    Wrapper around bzrlib.urlutils.unescape_for_display
    """
    if not url:
        return url
    return urlutils.unescape_for_display(url, 'utf-8')


def ActionList(sequence=None, prolog=None, epilog=None, subproject=None):
    """Return special list that holds sequence of actions plus prolog/epilog.

    Prolog/epilog code executed only once before/after processing components.
    They're suitable to implement pre/post action on control branch.
    """
    class ActionList(list):
        pass
    def check_item(item):
        if item is None:
            return []
        if getattr(item, '__iter__', None) is None:
            return [item]
        return item
    sequence = check_item(sequence)
    prolog = check_item(prolog)
    epilog = check_item(epilog)
    subproject = check_item(subproject)
    r = ActionList(sequence)
    r.prolog = prolog
    r.epilog = epilog
    r.subproject = subproject
    return r


class Action(object):
    """Base class for actions implementations.

    Instance should provide callable object that will be invoked
    in conext of each component and its branches.
    """

    def __init__(self):
        # I think you'll need to override constructor in your action
        pass

    def __call__(self, to_file, ko, kb, local_paths, local_urls):
        """Execute action for some component or component branch.

        @param  to_file:        stream to write messages.
        @param  ko:             Component object.
        @param  kb:             KBranch object.
        @param  local_paths:    dictionary with local paths
                                suitable for this component/branch
                                local storage.
        @param  local_urls:     URLs counterparts for local_paths
        @return: None if action executed successfully,
            otherwise return code will be used as return code for CLI.
        """
        raise NotImplementedError


class ActionV2(Action):
    """Base class for actions of projects V2."""

    def __call__(self, to_file, node, local_paths, local_urls):
        """Execute action for some component or component branch.

        @param  to_file:        stream to write messages.
        @param  node:           ProjectNode object.
        @param  local_paths:    dictionary with local paths
                                suitable for this component/branch
                                local storage.
        @param  local_urls:     URLs counterparts for local_paths
        @return: None if action executed successfully,
            otherwise return code will be used as return code for CLI.
        """
        raise NotImplementedError


class _CommandsAction(Action):
    """Run several commands with templates in arguments."""

    _show_kmd_prefix = ''

    def __init__(self, command_list, ignore_errors=False, verbose=True,
        dry_run=False, vars_dict=None):
        """Prepare action.
        @param command_list:    list of command templates
        @param ignore_errors:   don't stop on OSError/BzrError.
        @param verbose:         verbose comments what's going on
        @param dry_run:         show the command, but don't run it
        @param vars_dict:       additional dictionary with template variables
        """
        self.command_list = command_list
        self.ignore_errors = ignore_errors
        self.verbose = verbose
        self.dry_run = dry_run
        self.vars_dict = vars_dict or {}

    def __call__(self, to_file, ko, kb, local_paths, local_urls):
        engine = config.TemplateEngine({
                'RELPATH': ko.relpath,
                'BRANCH_NAME': kb.name,
                'BRANCH_URL': kb.url,
                'FORMAT': kb.format or 'default',
                'REVISION': kb.revision or '-1',
                })
        engine.add_from_dict(self.vars_dict)
        engine.add_from_dict(local_paths)
        for cmd_tmpl in self.command_list:
            kmd = engine.substitute(cmd_tmpl)    # create valid bzr command
            if self.verbose:
                to_file.write('#   Run command: %s%s\n' %
                    (self._show_kmd_prefix, kmd))
            if self.dry_run:
                continue
            return self._run_cmd(kmd)

    def _run_cmd(self, kmd):
        raise NotImplementedError


class BzrCommandsAction(_CommandsAction):
    """Run several bzr commands with templates in arguments."""

    _show_kmd_prefix = 'bzr '

    def _run_cmd(self, kmd):
        # shlex.split does not like back slashes :-(
        kmd = kmd.replace('\\', '/')
        argv = [p.decode('utf8') for p in shlex.split(kmd.encode('utf8'))]
        ret = run_bzr_catch_user_errors(argv)
        if ret == errors.EXIT_ERROR and not self.ignore_errors:
            return ret


class ShellCommandsAction(_CommandsAction):
    """Run several shell commands with templates in arguments."""

    def _run_cmd(self, kmd):
        ret = os.system(kmd)
        if ret != 0 and not self.ignore_errors:
            return ret


class _CommandsActionV2(Action):
    """Run several commands with templates in arguments."""

    _show_kmd_prefix = ''

    def __init__(self, command_list, ignore_errors=False, verbose=True,
        dry_run=False, vars_dict=None):
        """Prepare action.
        @param command_list:    list of command templates
        @param ignore_errors:   don't stop on OSError/BzrError.
        @param verbose:         verbose comments what's going on
        @param dry_run:         show the command, but don't run it
        @param vars_dict:       additional dictionary with template variables
        """
        self.command_list = command_list
        self.ignore_errors = ignore_errors
        self.verbose = verbose
        self.dry_run = dry_run
        self.vars_dict = vars_dict or {}

    def __call__(self, to_file, node, local_paths, local_urls):
        engine = config.TemplateEngine({
                'RELPATH': node.relpath,
                'BRANCH_NAME': node.kbranch.name,
                'BRANCH_URL': node.kbranch.url,
                'FORMAT': node.kbranch.format or 'default',
                'REVISION': node.kbranch.revision or '-1',
                })
        engine.add_from_dict(self.vars_dict)
        engine.add_from_dict(local_paths)
        for cmd_tmpl in self.command_list:
            kmd = engine.substitute(cmd_tmpl)    # create valid bzr command
            if self.verbose:
                to_file.write('#   Run command: %s%s\n' %
                    (self._show_kmd_prefix, kmd))
            if self.dry_run:
                continue
            return self._run_cmd(kmd)

    def _run_cmd(self, kmd):
        raise NotImplementedError


class BzrCommandsActionV2(_CommandsActionV2):
    """Run several bzr commands with templates in arguments."""

    _show_kmd_prefix = 'bzr '

    def _run_cmd(self, kmd):
        # shlex.split does not like back slashes :-(
        kmd = kmd.replace('\\', '/')
        argv = [p.decode('utf8') for p in shlex.split(kmd.encode('utf8'))]
        ret = run_bzr_catch_user_errors(argv)
        if ret == errors.EXIT_ERROR and not self.ignore_errors:
            return ret


class ShellCommandsActionV2(_CommandsActionV2):
    """Run several shell commands with templates in arguments."""

    def _run_cmd(self, kmd):
        ret = os.system(kmd)
        if ret != 0 and not self.ignore_errors:
            return ret


class CommitAction(Action):
    """Commit changes in the component working copy."""

    def __init__(self, msgfile):
        """@param msgfile: file with commit message."""
        self.msgfile = msgfile

    def __call__(self, to_file, ko, kb, local_paths, local_urls):
        note = to_file.write
        kb_vo = vcs_reg.get(kb.vcs)()
        path = local_paths['LOCAL_TREE']
        if not kb_vo.is_checkout(path):
            note('#   Skipped: no working copy.\n')
        else:
            note('#   Committing changes...\n')
            kb_vo.commit_tree(path, self.msgfile)


class CommitAndCollectRevidActionV2(Action):
    """Commit changes in the component working copy, and collect revids."""

    def __init__(self, msgfile, seed_revids=None):
        """@param msgfile: file with commit message.
        @param project_object: instance of Project for which commit executed.
        """
        self.msgfile = msgfile
        if seed_revids:
            self.revids = seed_revids
        else:
            self.revids = {}

    def __call__(self, to_file, node, local_paths, local_urls):
        if not node.is_subproject:
            # plain component
            note = to_file.write
            kb = node.kbranch
            kb_vo = vcs_reg.get(kb.vcs)()
            path = local_paths['LOCAL_TREE']
            if not kb_vo.is_checkout(path):
                note('#   Skipped: no working copy.\n')
            else:
                if node.is_root:
                    if not node.component.is_v1():
                        snapshot = project.SnapshotV2(node.abspath)
                        snapshot.update_index_card(self.revids)
                    self.revids = {}
                note('#   Committing changes...\n')
                kb_vo.commit_tree(path, self.msgfile)
                tree_url = local_urls['LOCAL_TREE']
                if not node.is_root:
                    self.revids[node.name] = kb_vo.get_tip_revid(tree_url)
        else:
            # subproject: run the action recursively
            subp = node.subproject
            if subp is not None:
                subtree = node.subtree
                revid = subp.commit_project_tree_v2(subtree, self.msgfile,
                    to_file)
                tree_url = local_urls['LOCAL_TREE']
                self.revids[node.name] = bzr_vcs.get_tip_revid(tree_url)


class UpdateAndCommitSnapshotEpilogV2(Action):

    def __init__(self, commit_action):
        """@param commit_action: CommitAndCollectRevidActionV2 instance from
            where we will get snapshot info.
        """
        self.commit_action = commit_action

    def __call__(self, to_file, local_paths, local_urls):
        revids = self.commit_action.revids
        if not revids:
            return
        root_path = local_paths['PROJECT_ROOT']
        # check that we're working with project V2
        if not osutils.isdir('.bzrmeta/scmproj'):   # ugly hack as workaround for Bug #563807
            return
        snapshot = project.SnapshotV2(root_path)
        if snapshot.update_index_card(revids):
            to_file.write('Updating snapshot of the project at ')
            to_file.write(root_path)
            to_file.write('\n')
            bzr_vcs.commit_tree(root_path, self.commit_action.msgfile, quiet=True)


class _CollectRevidsAction(Action):
    """Collect revid info for components.
    Used internally by Snapshot class.
    """
    def __init__(self):
        self.components = {}
        self.subprojects = {}

    def __call__(self, to_file, ko, kb, local_paths, local_urls):
        kb_vo = vcs_reg.get(kb.vcs)()
        path = local_paths['LOCAL_TREE']
        if kb_vo.is_checkout(path):
            tree_url = local_urls['LOCAL_TREE']
            self.components[ko.name] = kb_vo.get_tip_revid(tree_url)
                                            

class UpdateWorkingTreeAction(Action):
    """Update working tree of the component (plain ws).

    Pull new revisions to standalone branch in local project copy.
    """

    def __init__(self, create=True):
        self.create = create

    def __call__(self, to_file, ko, kb, local_paths, local_urls):
        note = to_file.write
        kb_vo = vcs_reg.get(kb.vcs)()
        path = local_paths['LOCAL_TREE']
        relpath = ko.relpath
        if relpath == '.':
            relpath = '"."'
        if not kb_vo.is_branch(path):
            if self.create:
                note('#   Creating working tree in %s...\n' % relpath)
                kb_vo.create_branch(path, kb.format, create_prefix=True)
            else:
                note('#   No working tree at %s\n' % relpath)
                return
        note('#   Updating working tree from %s...\n' % url_for_display(kb.url))
        kb_vo.pull(path, kb.url, kb.revision)


class PublishKBranch(Action):
    """Publish (push) component branch to origin."""

    def __init__(self, strict=None):
        self._strict = strict

    def __call__(self, to_file, ko, kb, local_paths, local_urls):
        note = to_file.write
        kb_vo = vcs_reg.get(kb.vcs)()
        local_branch = local_urls['LOCAL_BRANCH']
        if kb_vo.is_branch(local_branch):
            note('#   Pushing %s to %s...\n' % (kb.name, url_for_display(kb.url)))
            kb_vo.push(kb.url, local_branch, strict=self._strict)
        else:
            note('#   Skipped: no branch at %s\n' % local_paths['LOCAL_BRANCH'])
