# Simple support for organizing vcs branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

from bzrlib.commands import Command
from bzrlib.option import Option, ListOption, RegistryOption

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
import os
import re
import tempfile

from bzrlib import (
    branch,
    commands,
    errors,
    osutils,
    trace,
    transport,
    urlutils,
    )

from bzrlib.plugins.scmproj import (
    action,
    project,
    vcs,
    )
""")


def deprecated_option(option, name, new_value, parser):
    trace.warning('scmproj: WARNING: Option --%s is deprecated.' %
        name.replace('_','-'))

def _get_one_revision(command_name, revisions):
    if revisions is None:
        return None
    if len(revisions) != 1:
        raise errors.BzrCommandError(
            'bzr %s --revision takes exactly one revision identifier' % (
                command_name,))
    return revisions[0]


class cmd_project_init(Command):
    """Initialize new top-level branch for the project.

    Command will create .bzrmeta/scmproj directory in the tree
    and put template of project.cfg config there. In the case of
    remote location this config will be committed in the branch.
    In the case of local project you will need to commit it
    yourself.
    """
    aliases = ['pinit']
    takes_args = ['location']
    takes_options = []

    def run(self, location=None):
        project.Project.initialize(location)
        if project.is_url_local(location):
            if location == '.':
                directory = 'current directory'
            else:
                directory = '"%s"' % action.url_for_display(location)
            self.outf.write(('New top-level branch intialized in %s.\n'
                'Use command `project-add` to add components to your project,\n'
                'or edit %s in .bzrmeta/scmproj subdirectory '
                'to describe your project.\n')
                % (directory, project.CONFIG))
        else:
            self.outf.write(('New scmproj control branch intialized at "%s".\n'
                ) % action.url_for_display(location))


class cmd_project_add(Command):
    """Add new component or subproject to the current project.

    RELPATH is the component location relative to the project root.
    URL is the origin of component/subproject branch.

    Component name can be specified explicitly with --name option,
    otherwise it will be deduced from RELPATH.

    If you want to add another project as subproject you should supply
    --subproject option.
    """
    aliases = ['padd']
    takes_args = ['relpath', 'url']
    takes_options = [
        Option('name', type=unicode, help='Name of the component.'),
        Option('subproject', help='Add new subproject.'),
        ]

    def run(self, relpath, url, name=None, subproject=None):
        # XXX do we want the -d option to specify the project root?
        p = project.Project.open_containing('.')
        if p.format == 'v1':
            raise errors.BzrError('project-add command does not work for project V1.')
        root_path = p.t.local_abspath('.')
        abspath = p.t.local_abspath(relpath)
        try:
            the_relpath = osutils.relpath(root_path, abspath)
        except errors.PathNotChild:
            raise errors.BzrError('supplied RELPATH is outside of the project tree.')
        the_name = name
        if not name:
            the_name = urlutils.derive_to_location(the_relpath)
        class ComponentNameError(errors.BzrError):
            pass
        try:
            # validating the name
            try:
                the_name.encode('ascii')
            except UnicodeEncodeError:
                raise ComponentNameError('Name of component should be ASCII-only string.')
            # XXX add the helper method to config.py, e.g. validate_component_name
            #     see also https://bugs.launchpad.net/bzr-scmproj/+bug/458937
            if not p.cfg.validate_name(the_name, None, False):
                raise ComponentNameError('Invalid name of the component.')
            if the_name in p.cfg.components or the_name in p.cfg.subprojects:
                raise ComponentNameError('Project already has the component/subproject with name "%s".' % the_name)
        except ComponentNameError, e:
            if not name:
                raise errors.BzrError('Please, provide explicit ASCII-only name for the component. (%s)' % str(e))
            raise
        #
        for kname in p.cfg.components:
            if p.cfg.get_component_relpath(kname) == the_relpath:
                raise errors.BzrError('Project already has the component at relpath "%s".' % the_relpath)
        for sname in p.cfg.subprojects:
            if p.cfg.get_subproject_relpath(sname) == the_relpath:
                raise errors.BzrError('Project already has the subproject at relpath "%s".' % the_relpath)
        #
        add_func = p.cfg.add_component
        kind = 'Component'
        if subproject:
            add_func = p.cfg.add_subproject
            kind = 'Subproject'
        add_func(the_name, the_relpath, url)
        p.save_config()
        p.ignore_relpath(the_relpath)
        trace.info('%s "%s" added successfully.' % (kind, the_name))
        trace.note('To get the local copy of the component use `bzr project-update` command.')


class cmd_project_branch(Command):
    """Change the origin of the component.
    """
    aliases = ['pbranch']
    takes_args = ['new_url']
    takes_options = [
        Option('component', short_name='k', type=unicode,
            help="Component name."),
        Option('root', help='Perform operation on root component.'),
        Option('push', help='Push local branch to new origin.'),
        Option('pull', help='Update local branch from new origin.'),
        ]

    def run(self, new_url, component=None, root=False, push=False, pull=False):
        p = project.Project.open_containing('.')
        if p.format == 'v1':
            raise errors.BzrError('project-branch command does not work for project V1.')
        if component is None and not root:
            raise errors.BzrError('Please provide component name via --component option.')
        if component and root:
            raise errors.BzrError("You can't use --root and --component in the same time.")
        if root:
            relpath = '.'
            # [bialix 2010/05/01] maybe I'll finish it later when I really will need this
            ## check for surrounding mega project
            #parent_dir = p.t.local_abspath('..')
            #try:
            #    mp = project.Project.open_containing(parent_dir)
            #except errors.BzrError:
            #    pass
            #else:
            #    subp_relpath = osutils.relpath(mp.t.local_abspath('.'),
            #        p.t.local_abspath('.'))
            #    subp = mp.cfg.get_by_relpath(subp_relpath)
            #    if subp:
            #        XXX
        else:
            p.cfg.change_origin(component, new_url)
            p.save_config()
            relpath = p.cfg.get_relpath_of(component)
        t = transport.get_transport(new_url)
        abspath = p.t.local_abspath(relpath)
        try:
            br = branch.Branch.open(abspath)
        except errors.NotBranchError:
            pass
        else:
            br.set_parent(t.base)
            if br.get_push_location():
                br.set_push_location(t.base)
            if push:
                vcs.bzr_vcs.push(new_url, relpath)
                if root:
                    trace.note('Run `project-publish` if you want to publish the entire project.')
        if pull:
            if not root:
                vcs.bzr_vcs.pull(relpath, new_url)
            else:
                commands.run_bzr('project-update')


class cmd_project_info(Command):
    """Show info summary for project."""
    aliases = ['pinfo']
    takes_args = ['location?']
    takes_options = []

    def run(self, location=None):
        if location is None:
            location = '.'
        p = project.Project.open_containing(location)
        p.info(self.outf)


class cmd_project_get(Command):
    """Get a copy of an existing project on your local disk.

    This will make a new project dir with a .scmproj control branch inside,
    and get component branches of the project.

    FROM_LOCATION may refer to a local project dir or a remote scmproj
    control branch.
    If the TO_LOCATION is omitted, the last component of the FROM_LOCATION
    will be used.  In other words, "project-get ../my/proj" will attempt to
    create ./proj.
    """
    aliases = ['pget']
    takes_args = ['from_location?', 'to_location?']
    takes_options = [
        'revision',
        Option('control-branch-only',
            help="Only get the control branch (i.e. don't get the component branches).",
            hidden=True, custom_callback=deprecated_option),
        Option('components-only', help="Only get the component branches "
            "(when the control branch is already present).",
            hidden=True, custom_callback=deprecated_option),
        Option('alt', type=unicode, help='Get branches for the specified alt. (deprecated)',
            custom_callback=deprecated_option),
        Option('subset', type=unicode, help='Get components of the specified subset.',
            hidden=True, custom_callback=deprecated_option),
        Option('full-set', help='Get all components.',
            hidden=True, custom_callback=deprecated_option),
        ]

    def run(self, from_location=None, to_location=None, revision=None,
        alt=None, subset=None,
        full_set=None, control_branch_only=None, components_only=None):
        revision = _get_one_revision('project-get', revision)
        if control_branch_only and components_only:
            raise errors.BzrError("You cannot use both --control-branch-only "
                "and --components-only options in the same time.")
        if not components_only:
            if from_location is None:
                raise errors.BzrError(
                    'You should specify source location as first argument.')
            # clone control branch first
            if to_location is None:
                to_location = urlutils.derive_to_location(from_location)
            project.Project(from_location).clone(to_location, revision=revision)
        else:
            # control branch already should be cloned
            if (from_location, to_location) != (None, None):
                raise errors.BzrError('To get only components branches '
                    'run project-get command without arguments.')
            to_location = '.'

        p = project.Project(to_location).open()

        class GetCommand(project.ProjectCommand):
            def __init__(self, main_alt, main_component_list, **kw):
                super(GetCommand, self).__init__(**kw)
                self.main_alt = main_alt
                self.main_component_list = main_component_list
                self.clone_subproject = True

            def execute(self, p):
                if p.i_am_subproject:
                    alt, component_list = p.cli_helper()
                else:
                    alt = self.main_alt
                    component_list = self.main_component_list
                return p.update_workspace(alt_name=alt,
                    component_list=component_list,
                    **self.kw)

        ret_code = None
        if not control_branch_only:
            alt, component_list = p.cli_helper(alt=alt, subset=subset, full_set=full_set)
            get_cmd = GetCommand(alt, component_list,
                to_file=self.outf)  # kw
            ret_code = get_cmd.execute(p)
        p.save_settings(alt, subset)
        return ret_code


# A special option so 'revision' can be passed as a simple string, when we do
# *not* wan't bzrlib's feature of parsing the revision string before passing it.
# This is used when we just want a plain string to pass down to vcs layer.
# If you want bzrlib to parse and pass a revisionspec object, just pass the string
# 'revision' as normal.
simple_revision_option = Option("revision",
                             short_name='r',
                             type=unicode,
                             help='See "help revisionspec" for details.')


class cmd_project_update(Command):
    """Update components of the project from their origins."""

    aliases = ['pup', 'pupdate']
    takes_args = ['location?']
    takes_options = [
        simple_revision_option,
        Option('components-only', help="Update only component branches.",
            hidden=True, custom_callback=deprecated_option),
        Option('all', short_name='A', help='Force update all components.'),
        ]

    def run(self, location=None, revision=None, components_only=None, all=False):
        if location is None:
            location = '.'
        p = project.Project.open_containing(location)

        class UpdateCommand(project.ProjectCommand):
            def __init__(self, components_only, to_file, rev_str, **kw):
                super(UpdateCommand, self).__init__(**kw)
                self.components_only = components_only
                self.to_file = to_file
                self.rev_str = rev_str
                self.clone_subproject = True

            def execute(self, p):
                if not self.components_only:
                    p.update_control_branch(self.to_file, rev_str=self.rev_str)
                    self.rev_str = None
                alt, component_list = p.cli_helper()
                return p.update_workspace(to_file=self.to_file,
                    alt_name=alt, component_list=component_list,
                    **self.kw)

        update_cmd = UpdateCommand(components_only=components_only,
            to_file=self.outf, rev_str=revision,
            all=all)    # kw
        return update_cmd.execute(p)


class cmd_project_publish(Command):
    """Publish (push back) changes from components branches and control branch."""
    aliases = ['pback', 'project-push', 'ppush']
    takes_args = ['component*']
    takes_options = [
        Option('alt', argname='alt', type=unicode, help='Push data for specified alt.',
            hidden=True, custom_callback=deprecated_option),
        Option('subset', type=unicode, help='Push data for specified subset.',
            hidden=True, custom_callback=deprecated_option),
        Option('full-set', help='Push data for all components.',
            hidden=True, custom_callback=deprecated_option),
        Option('components-only', help="Publish only component branches.",
            hidden=True, custom_callback=deprecated_option),
        Option('strict',
               help='Refuse to push if there are uncommitted changes in'
               ' components trees, --no-strict disables the check.'),
        Option('all', short_name='A', help='Force publish all components.'),
        ]

    def run(self, component_list=None, alt=None, subset=None, full_set=None,
        components_only=None, strict=None, all=False):
        p = project.Project.open_containing('.')

        class PublishCommand(project.ProjectCommand):
            def __init__(self, component_list, alt, subset, full_set,
                components_only, **kw):
                super(PublishCommand, self).__init__(**kw)
                self.component_list = component_list
                self.alt = alt
                self.subset = subset
                self.full_set = full_set
                self.components_only = components_only

            def execute(self, p):
                if not p.i_am_subproject:
                    alt, klist = p.cli_helper(alt=self.alt, subset=self.subset,
                        full_set=self.full_set)
                    component_list = self.component_list
                    if component_list is None:
                        component_list = klist
                else:
                    alt, component_list = p.cli_helper()
                return p.publish(alt_name=alt, component_list=component_list,
                    components_only=self.components_only, **self.kw)

        publish_cmd = PublishCommand(component_list, alt, subset, full_set,
            components_only,
            to_file=self.outf, strict=strict, all=all)  # kw
        return publish_cmd.execute(p)


class cmd_project_command(Command):
    """Run some command(s) for project components.

    .. note::
    
       Run this command from the root of local project copy.

    Option --cd-relpath (-C) will force execution of command(s) in the
    components working directories. It's equal to have "cd {RELPATH}"
    prefixed to each command. This is default behavior. If you need
    to suppress automatic directory change, please use --no-cd option.

    If you want to check the actual commands and its arguments for every
    component without actually invoking these commands, then you may run
    with options --dry-run and --verbose (-v), e.g::

        bzr pcmd -v --dry-run "missing {BRANCH_URL}"
    """
    aliases = ['pcmd', 'prun', 'pcommand']
    takes_args = ['command+']
    takes_options = [
        'verbose',
        Option('dry-run',
            help="Show what would be done, but don't actually do anything."),
        Option('ignore-errors', short_name='E',
            help="Don't stop on user errors."),
        Option('alt', argname='alt', type=unicode,
            help='Run command for specified alt.',
            hidden=True, custom_callback=deprecated_option),
        Option('subset', type=unicode,
            help='Run command for specified subset.',
            hidden=True, custom_callback=deprecated_option),
        Option('full-set', help='Run command for all components.',
            hidden=True, custom_callback=deprecated_option),
        Option('cd-relpath', short_name='C',
            help="Cd to component working directory before executing commands."),
        Option('shell', short_name='s',
            help='Run shell commands instead of bzr commands.'),
        Option('var1', type=unicode,
            help='Additional template variable {var1}.'),
        Option('var2', type=unicode,
            help='Additional template variable {var2}.'),
        Option('var3', type=unicode,
            help='Additional template variable {var3}.'),
        ListOption('include', short_name='i', type=unicode,
            help="Run command only for specified components."),
        ListOption('exclude', short_name='x', type=unicode,
            help="Don't run command for specified components."),
        RegistryOption('select',
            help="Run command only for components which conform the status.",
            value_switches=True,
            lazy_registry=('bzrlib.plugins.scmproj.vcs', 'checkers_reg'),
            ),
        ]

    def run(self, command_list, ignore_errors=False, alt=None, subset=None,
        full_set=None, cd_relpath=True, verbose=False, dry_run=False,
        shell=False, var1=None, var2=None, var3=None,
        include=None, exclude=None, changed=None, select=None):
        if include and exclude:
            raise errors.BzrError("You can't use --include and --exclude "
                "in the same time.")
        p = project.Project.open_containing('.')
        alt, component_list = p.cli_helper(alt=alt, subset=subset,
                                           full_set=full_set)
        vars_dict = p.vars_helper(var1=var1, var2=var2, var3=var3)
        if p.format == 'v2':
            return p.run_cmds_v2(command_list, to_file=self.outf,
                ignore_errors=ignore_errors,
                includes=include, excludes=exclude,
                cd_relpath=cd_relpath, verbose=verbose,
                dry_run=dry_run, shell=shell, vars_dict=vars_dict,
                checker_func=select)
        # else v1
        if include:
            new_component_list = []
            for k in component_list:
                if k in include:
                    new_component_list.append(k)
            component_list = new_component_list
        if exclude:
            new_component_list = []
            for k in component_list:
                if k not in exclude:
                    new_component_list.append(k)
            component_list = new_component_list

        class PCmdCommand(project.ProjectCommand):
            def __init__(self, main_alt, main_component_list, **kw):
                super(PCmdCommand, self).__init__(**kw)
                self.main_alt = main_alt
                self.main_component_list = main_component_list

            def execute(self, p):
                if p.i_am_subproject:
                    alt, component_list = p.cli_helper()
                else:
                    alt = self.main_alt
                    component_list = self.main_component_list
                return p.run_cmds(alt_name=alt, components=component_list,
                    **self.kw)

        pcmd_cmd = PCmdCommand(alt, component_list,
            command_list=command_list, ignore_errors=ignore_errors,  # kw
            to_file=self.outf, cd_relpath=cd_relpath, verbose=verbose,
            dry_run=dry_run, shell=shell, vars_dict=vars_dict,
            checker_func=select)
        return pcmd_cmd.execute(p)


class cmd_chdir(Command):
    """Run bzr command in another directory.

    This command changes the current directory to another one and then
    launches the bzr command provided as the argument, e.g.::

        bzr chdir foo/bar -- missing http://example.com/spam

    This is equivalent to::

        # remember $CWD
        cd foo/bar
        bzr missing http://example.com/spam
        cd $ORIG_CWD

    This is ad-hoc solution to workaround some bugs or missing
    features in core bzr commands for batch processing.
    Intended to use with command `project-command`.

    .. note::
    
       You always should use ``--`` after directory you want to cd
       and before actual subcommand you want to execute in that directory.
       Otherwise it's not possible to pass options to the subcommand, e.g.::

        bzr chdir foo/bar -- missing --line http://example.com/spam

       This command will not work without ``--`` between ``foo/bar``
       and ``missing``.
    """
    aliases = ['cd']
    takes_args = ['directory', 'argv+']
    takes_options = []
    hidden = True

    def run(self, directory, argv_list):
        trace.warning('This command is deprecated and will be removed in the next release. '
            "Don't use it.")
        from bzrlib.commands import run_bzr
        cwd = os.getcwdu()
        os.chdir(directory)
        try:
            return run_bzr(argv_list)
        finally:
            os.chdir(cwd)


class cmd_project_commit(Command):
    """Commit changes in components working copies."""
    aliases = ['pci', 'pcommit']
    takes_args = ['component*']
    takes_options = [
        Option('subset', type=unicode, help='Commit components of specified subset.',
            hidden=True, custom_callback=deprecated_option),
        Option('full-set', help='Commit all components.',
            hidden=True, custom_callback=deprecated_option),
        Option('file', type=str, short_name='F', argname='msgfile',
               help='Take commit message from this file.'),
        Option('message', type=unicode, short_name='m',
               help="Commit message."),
        Option('components-only', help="Commit only changes in components.",
            hidden=True, custom_callback=deprecated_option),
        ]

    def run(self, component_list=None, subset=None, full_set=None, file=None,
        message=None, components_only=None):
        if file is None and not message:
            raise errors.BzrError('You should provide commit message.')
        elif message and file:
            raise errors.BzrCommandError("please specify either --message or --file")
        if message:     # save commit message to temp file
            tmp_fileno, file = tempfile.mkstemp()
            msgfile = os.fdopen(tmp_fileno, 'w')
            try:
                msgfile.write(message.encode(osutils.get_user_encoding()))
            finally:
                msgfile.close()
        try:
            p = project.Project.open_containing('.')
            if p.format == 'v2':
                return p.commit_v2(file, self.outf, component_list)

            # else v1
            if component_list is None:
                component_list = p.get_component_list(subset, full_set)

            class CommitCommand(project.ProjectCommand):
                def __init__(self, msgfile, main_component_list, **kw):
                    super(CommitCommand, self).__init__(**kw)
                    self.msgfile = msgfile
                    self.main_component_list = main_component_list

                def execute(self, p):
                    if p.i_am_subproject:
                        component_list = p.get_component_list()
                    else:
                        component_list = self.main_component_list
                    return p.commit(self.msgfile, component_list=component_list,
                        **self.kw)

            commit_cmd = CommitCommand(file, component_list,
                to_file=self.outf, components_only=components_only)  # kw
            return commit_cmd.execute(p)
        finally:
            if message:     # if message have been saved we need to delete temp file
                try:
                    os.unlink(file)
                except IOError, e:
                    trace.warning("failed to unlink %s: %s; ignored", file, e)


class cmd_dump_ptree(Command):
    """Dump the ProjectTree."""
    hidden = True
    takes_options = [Option('snode'), Option('no-recurse'), Option('reverse'),
        ListOption('include', short_name='i', type=unicode,
            help="Run command only for specified components."),
        ListOption('exclude', short_name='x', type=unicode,
            help="Don't run command for specified components."),
        ]
    def run(self, snode=False, no_recurse=False, reverse=False,
        include=None, exclude=None):
        p = project.Project.open_containing('.')
        ptree = project.ProjectTree(p, True)
        ptree.build_tree()
        if not reverse:
            nodes = list(ptree.iterate_filter(snode, not no_recurse, include, exclude))
        else:
            nodes = list(ptree.iterate_filter_reverse(snode, not no_recurse, include, exclude))
        from bzrlib import osutils
        import pprint
        pprint.pprint(nodes)
