# Simple support for organizing vcs branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

from cStringIO import StringIO

from bzrlib import (
    errors,
    urlutils,
    )
import bzrlib.config
from bzrlib.lazy_regex import lazy_compile
# configobj library
try:
    from bzrlib.util.configobj import configobj
except ImportError:
    import configobj

from bzrlib.plugins.scmproj.pythonutils import odict

# See project.cfg and local.cfg skeletons at the end of file
# global variables CONFIG_TMPL and LOCAL_TMPL


def list_from_str(s):
    return filter(None, s.strip().split())


class TemplateEngine(object):

    _tag_re = lazy_compile('{(\w+)}')

    def __init__(self, dikt=None, **kw):
        self._data = {}
        if dikt is not None:
            if isinstance(dikt, dict):
                self._data.update(dikt)
            else:
                raise TypeError('initializer should be a dict')
        if kw:
            self._data.update(kw)

    def add(self, name, value):
        self._data[name] = value

    def add_from_dict(self, dikt):
        """Add template variables from dictionary."""
        self._data.update(dikt)

    def substitute(self, tpl):
        """Process template and substitute {names} with their values.
        @param  tpl:    possible template string.
        @return:    string after processing.
        """
        res = []
        pos = 0
        while True:
            match = self._tag_re.search(tpl, pos)
            if not match:
                if pos < len(tpl):
                    res.append(tpl[pos:])
                break
            start, end = match.span()
            if start > 0:
                res.append(tpl[pos:start])
            pos = end
            name = match.group(1)
            try:
                data = self._data[name]
            except KeyError:
                raise errors.MissingTemplateVariable(name)
            if not isinstance(data, basestring):
                data = str(data)
            res.append(data)
        return ''.join(res)


class ConfigObjEx(configobj.ConfigObj):

    @staticmethod
    def from_str(s):
        return ConfigObjEx(StringIO(s), encoding='utf-8')

    def to_str(self):
        sio = StringIO()
        self.write(sio)
        return sio.getvalue()

    def _write_marker(self, indent_string, depth, entry, comment):
        """Write a section marker line"""
        return '%s%s%s%s%s' % (indent_string,
                               self._a_to_u('[' * depth),
# [bialix 20081218]
# we don't want to have section names with spaces in it (e.g. [component "foo"]) quoted
# ConfigObj itself has no problem to parse it from file.
# do you think it will be unsafe?
#                               self._quote(self._decode_element(entry), multiline=False),
                               self._decode_element(entry),
#
                               self._a_to_u(']' * depth),
                               self._decode_element(comment))


class LocalSettings(object):
    """Local settings (for new layout)."""

    def __init__(self, control_branch):
        """@param control_branch: opened branch in the .scmproj control dir"""
        self.control_branch = control_branch
        self.cfg = control_branch.get_config()

    def _get_project_url(self):
        return self.control_branch.get_parent()

    def _set_project_url(self, url):
        # XXX why for we ever need to change this parameter???
        return self.control_branch.set_parent(url)

    project_url = property(_get_project_url, _set_project_url)

    def _get_alt(self):
        r = self.cfg.get_user_option('scmproj:alt')
        if r is None:
            return self.cfg.get_user_option('scmproj_alt')
        return r

    def _set_alt(self, alt):
        self.cfg.set_user_option('scmproj:alt', alt)

    alt = property(_get_alt, _set_alt, doc="Current alt.")

    def _get_subset(self):
        r = self.cfg.get_user_option('scmproj:subset')
        if r is None:
            return self.cfg.get_user_option('scmproj_subset')
        return r

    def _set_subset(self, subset):
        self.cfg.set_user_option('scmproj:subset', subset)

    subset = property(_get_subset, _set_subset, doc="Current subset.")

    def _get_ws_name(self):
        ws = self.cfg.get_user_option('scmproj:workspace')
        if ws == 'boxed':
            raise errors.BzrError("boxed workspaces do not supported anymore.")
        return ws

    def _set_ws_name(self, ws_name):
        raise errors.BzrError("save workspace name does not supported anymore.")

    workspace = property(_get_ws_name, _set_ws_name)


class LocalSettingsV2(LocalSettings):
    """Handle settings for v2 layout: project.cfg stored in .bzrmeta/scmproj"""

    def _get_alt(self):
        return None

    def _set_alt(self, alt):
        raise errors.BzrError("set alt does not supported in v2.")

    alt = property(_get_alt, _set_alt, doc="Deprecated. Not used.")

    def _get_subset(self):
        return None

    def _set_subset(self, subset):
        raise errors.BzrError("set subset does not supported in v2.")

    subset = property(_get_subset, _set_subset, doc="Deprecated. Not used.")

    def _get_ws_name(self):
        raise errors.BzrError("read workspace name does not supported anymore.")

    def _set_ws_name(self, ws_name):
        raise errors.BzrError("save workspace name does not supported anymore.")

    workspace = property(_get_ws_name, _set_ws_name, doc="Deprecated. Not used.")


class BadComponentName(errors.BzrError):
    pass


class Config(object):
    """Project configuration."""

    VER = '### project.cfg v1'
    _warning = "### Don't edit the line above!"
    _ver_re = lazy_compile('^#+\s+project.cfg\s+v\.?1$')

    _name_re = lazy_compile('^[a-zA-Z][0-9a-zA-Z_\-]*$')
    _component_re = lazy_compile('^component[ \\t]+("?)([^"]*)("?)$')
    _branch_re = lazy_compile('^branch[ \\t]+("?)([^"]*)("?)$')
    _subproject_re = lazy_compile('^subproject[ \\t]+("?)([^"]*)("?)$')

    def __init__(self, s=None):
        if s is not None:
            self.from_str(s)
        else:
            self._defaults()

    def _defaults(self):
        self._cfg = None
        self.name = None
        self.description = None
        self.comp_sect = {}     # map between component name and corresponding section name in config file
        self.branch_sect = {}   # map for every component -> dict of branches sections
        self.subp_sect = {}     # map between subproject name and corresponding section name in config file
        self.var_sect = {}      # map for every subproject -> dict of variants sections
        self.vcs = None

    def from_str(self, s):
        self._defaults()
        if s:
            fmt_line = s.splitlines()[0].strip()
            if self._ver_re.match(fmt_line) is None:
                # TODO: suggest way to fix this situation
                # TODO: when we'll be close to format v2 we need to raise error
                from bzrlib import trace
                trace.warning('WARNING: '
                    'First line of project.cfg has no version marker')
        self._cfg = ConfigObjEx.from_str(s)
        sect = self._cfg.get('PROJECT', {})
        self.name = sect.get('NAME')
        self.description = sect.get('DESCRIPTION')
        self.comp_sect = None       # lazy collecting components, see _get_component_list below
        self.vcs = sect.get('VCS', 'bzr')
        self.component_templates = self._cfg.get('COMPONENTS', {})
        self.subp_sect = None       # lazy collecting subprojects, see _get_subproject_list below

    def _get_component_list(self):
        if self.comp_sect is None:
            implicit = self._collect_components()
            explicit = self._cfg.get('PROJECT', {}).get('COMPONENTS')
            if explicit:
                explicit = odict.OrderedDict(
                    [(i,None) for i in list_from_str(explicit)])
                if implicit:
                    # check that implicit is subset of explicit
                    if not set(explicit).issuperset(set(implicit)):
                        raise errors.BzrError(
                            'Explicit COMPONENTS list and '
                            'available components sections does not match.')
                explicit.update(implicit)
                self.comp_sect = explicit
            else:
                self.comp_sect = implicit
        return self.comp_sect.keys()

    components = property(_get_component_list,
        doc="list of project components names.")

    def to_str(self):
        if self._cfg is None:
            return '\n'.join((self.VER, self._warning, '', CONFIG_TMPL))
        return self._cfg.to_str()

    def get_subset_default_link(self):
        """Resolve link and return link target or 'DEFAULT'.
        E.g. for the config like this

        [SUBSETS]
        DEFAULT = @foo
        foo = ...

        The function will return 'foo'.
        """
        s = self._cfg.get('SUBSETS', {}).get('DEFAULT', '')
        if s.startswith('@'):
            return s[1:]
        return 'DEFAULT'

    def get_subset(self, subset):
        sv = self._cfg.get('SUBSETS', {})
        if subset == 'FULL':
            return self.get_full_set()
        elif subset == 'NULL':
            return []
        elif subset == 'DEFAULT':
            if sv == {} or 'DEFAULT' not in sv.scalars:
                return self.get_full_set()
            subset = self.get_subset_default_link()
        if subset not in sv:
            raise errors.BzrError("Could not find subset: %s" % alt)
        return self.sort_components(list_from_str(sv[subset]))

    def get_full_set(self):
        return self.sort_components(self.components)

    def sort_components(self, components):
        return sorted(components, key=self.get_component_relpath)

    def get_alt_default_link(self):
        """Resolve link and return link target or 'DEFAULT'.
        E.g. for the config like this

        [ALTS]
        DEFAULT = @foo
        foo = ...

        The function will return 'foo'.
        """
        s = self._cfg.get('ALTS', {}).get('DEFAULT', '')
        if s.startswith('@'):
            return s[1:]
        return 'DEFAULT'

    def get_alt(self, alt):
        sa = self._cfg.get('ALTS', {})
        if alt == 'DEFAULT':
            if sa == {} or 'DEFAULT' not in sa.scalars:
                return dict([(i, 'DEFAULT') for i in self.components])
            alt = self.get_alt_default_link()
        if alt not in sa:
            raise errors.BzrError("alt not found: %s" % alt)
        x = dict([i.split(':',1) for i in list_from_str(sa[alt])])
        default = x.get('{COMPONENT}', 'DEFAULT')
        r = {}
        for i in self.components:
            r[i] = x.get(i, default)
        return r

    def get_component_section(self, name):
        if self.comp_sect is None:
            self._get_component_list()
        sect_name = self.comp_sect[name]
        if sect_name is not None:
            return self._cfg[sect_name]
        else:
            return {}

    def _get_component_value(self, key, component_name, mandatory=True):
        cs = self.get_component_section(component_name)
        #XXX: we could swap these around
        if key in cs:
            value = cs[key]
        elif key in self.component_templates:
            value = self.component_templates[key]
        else:
            if mandatory:
                raise errors.BzrError(
                    "Could not determine %s for component %s" % (
                        key, component_name))
            else:
                return None
        return value

    def get_branch_section(self, component_name, branch_name):
        d = self.get_branch_sect_dict(component_name)
        sect_name = d.get(branch_name)
        if sect_name:
            return self.get_component_section(component_name)[sect_name]
        return None

    def _get_component_branch_value(self, key, component_name, branch_name):
        cs = self.get_component_section(component_name)
        branch_section = self.get_branch_section(component_name, branch_name)
        if branch_section is not None and key in branch_section:
            value = branch_section[key]
        elif key in cs:
            value = cs[key]
        elif key in self.component_templates:
            value = self.component_templates[key]
        else:
            raise errors.BzrError(
                "Could not determine %s for component branch %s:%s " % (
                    key, component_name, branch_name))
        return value

    def get_component_relpath(self, component_name):
        t = TemplateEngine({'COMPONENT': component_name})
        return t.substitute(
            self._get_component_value('RELPATH', component_name))

    def get_component_branches(self, component_name):
        explicit = None
        branches = self._get_component_value('BRANCHES', component_name,
            mandatory=False)
        if branches is not None:
            explicit = list_from_str(branches)
        d = self.get_branch_sect_dict(component_name)
        implicit = d.keys()
        if explicit is not None:
            if implicit:
                # check that implicit is subset of explicit
                if not set(explicit).issuperset(set(implicit)):
                    raise errors.BzrError(('Component %s: '
                        'explicit BRANCHES list and '
                        'available branches subsections does not match.'
                        ) % component_name)
            return explicit
        else:
            if implicit:
                return implicit
            # looking for valid BRANCH_URL settings
            url = self.get_component_branch_url(component_name, None)
            branch_name = urlutils.derive_to_location(url)
            return [branch_name]

    def get_component_branch_url(self, component_name, branch_name):
        t = TemplateEngine({
            'COMPONENT': component_name,
            'RELPATH' : self.get_component_relpath(component_name),
            #'ALT' not allowed, since a branch can belong to multiple alts.
            })
        if branch_name:
            t.add('BRANCH', branch_name)
        return t.substitute(
            self._get_component_branch_value('BRANCH_URL', component_name,
                branch_name))

    def ensure_components(self, components):
        """Check items in components list for presence in project
        known components list.
        @return:    components if all items is known components.
        @raise BzrError:    if there is unknown items.
        """
        for k in components:
            if k not in self.components:
                raise errors.BzrError('Unknown component: '+k)
        return components

    def validate_name(self, name, section_name, raise_error=True):
        if not self._name_re.match(name):
            if raise_error:
                raise BadComponentName('Bad name in section %r' % section_name)
            return False
        return True

    def _collect_components(self):
        d = odict.OrderedDict()
        for i in self._cfg.sections:
            # XXX ConfigObj seems to strip leading and trailing spaces from section names.
            mo = self._component_re.match(i)
            if mo:
                if mo.group(1) != mo.group(3):
                    raise errors.BzrError('Unmatched quotes in component '
                        'section: %r' % i)
                name = mo.group(2)
                self.validate_name(name, i)
                d[name] = i
        return d

    def _collect_branches(self, component_name):
        sect = self.get_component_section(component_name)
        d = odict.OrderedDict()
        for i in getattr(sect, 'sections', []):
            mo = self._branch_re.match(i)
            if mo:
                if mo.group(1) != mo.group(3):
                    raise errors.BzrError('Unmatched quotes in branches '
                        'section: %r' % i)
                name = mo.group(2)
                self.validate_name(name, i)
                d[name] = i
            else:
                raise errors.BzrError(('Unknown subsection "%s" in section '
                    'of component "%s"') % (i, component_name))
        return d

    def get_branch_sect_dict(self, component_name):
        d = self.branch_sect.get(component_name)
        if d is None:
            d = self._collect_branches(component_name)
            self.branch_sect[component_name] = d
        return d

    def get_default_format(self):
        """Return format value defined in COMPONENTS section or None."""
        return self._cfg.get('COMPONENTS', {}).get('FORMAT')

    def get_default_vcs(self):
        r = self._cfg.get('COMPONENTS', {}).get('VCS')
        if r is None:
            r = self._cfg.get('PROJECT', {}).get('VCS')
        if r is None:
            return 'bzr'
        return r

    def _collect_subprojects(self):
        d = odict.OrderedDict()
        for i in self._cfg.sections:
            # XXX ConfigObj seems to strip leading and trailing spaces from section names.
            mo = self._subproject_re.match(i)
            if mo:
                if mo.group(1) != mo.group(3):
                    raise errors.BzrError('Unmatched quotes in subproject '
                        'section: %r' % i)
                name = mo.group(2)
                self.validate_name(name, i)
                d[name] = i
        return d

    def _get_subproject_list(self):
        if self.subp_sect is None:
            self.subp_sect = self._collect_subprojects()
        # check that names of components and subprojects don't overlap
        cset = set(self.components)
        sset = set(self.subp_sect.keys())
        iset = cset.intersection(sset)
        if iset != set():
            raise errors.BzrError('Some subprojects have the names identical '
                'to components names: %s' % sorted(list(iset)))
        return sorted(self.subp_sect.keys())

    subprojects = property(_get_subproject_list,
        doc="list of subprojects names.")

    def get_subproject_section(self, name):
        if self.subp_sect is None:
            self.subp_sect = self._collect_subprojects()
        sect_name = self.subp_sect[name]
        return self._cfg[sect_name]

    def get_subproject_url(self, name):
        try:
            return self.get_subproject_section(name)['SUBPROJECT_URL']
        except KeyError, e:
            raise errors.BzrError(
                'Subproject "%s" lacks mandatory SUBPROJECT_URL' % name)

    def get_subproject_relpath(self, name):
        try:
            return self.get_subproject_section(name)['RELPATH']
        except KeyError, e:
            raise errors.BzrError(
                "Subproject %r section lacks mandatory RELPATH: %s",
                self.name, str(e))

    def sort_subprojects(self, subprojects=None):
        if subprojects is None:
            subprojects = self.subprojects
        return sorted(subprojects, key=self.get_subproject_relpath)

    def add_component(self, name, relpath, url):
        d = odict.OrderedDict()
        d['RELPATH'] = relpath
        d['BRANCH_URL'] = url
        self._cfg['component "%s"' % name] = d

    def add_subproject(self, name, relpath, url):
        d = odict.OrderedDict()
        d['RELPATH'] = relpath
        d['SUBPROJECT_URL'] = url
        self._cfg['subproject "%s"' % name] = d

    def change_origin(self, name, new_url):
        if name in self.components:
            d = self.get_component_section(name)
            d['BRANCH_URL'] = new_url
        elif name in self.subprojects:
            d = self.get_subproject_section(name)
            d['SUBPROJECT_URL'] = new_url
        else:
            raise errors.BzrError('No component with name "%s".' % name)

    def get_relpath_of(self, name):
        if name in self.components:
            return self.get_component_relpath(name)
        elif name in self.subprojects:
            return self.get_subproject_relpath(name)
        else:
            raise errors.BzrError('No component with name "%s".' % name)


class IndexCard(object):
    """Index card for snapshot.

    Index card holds info about revid for all components
    Index card has ini-style format with one (default) section:

        component_name = revid
    """

    VER = "# index card v2"

    def __init__(self, s=None):
        if s:
            self.from_str(s)
        else:
            self._cfg = ConfigObjEx.from_str(self.VER+'\n')

    def from_str(self, s):
        if not s:
            return
        first = s.splitlines()[0].rstrip()
        if first != self.VER:
            raise errors.BzrError('Wrong version of index card: '+repr(first))
        self._cfg = ConfigObjEx.from_str(s)

    def to_str(self):
        return self._cfg.to_str()

    def get_component(self, component_name):
        return self._cfg.get(component_name)

    def get_revision(self, component_name):
        r = self.get_component(component_name)
        if r:
            return 'revid:' + r
        return None

    def get_revid(self, component_name):
        r = self.get_component(component_name)
        if r:
            return r.encode('utf-8')
        return None

    def set_component(self, component_name, revid):
        self._cfg[component_name] = revid

    def update_revids(self, revids):
        if revids is None:
            return False
        result = False
        new_cfg = ConfigObjEx.from_str(self.VER+'\n')
        keys = sorted(revids.keys() + self._cfg.scalars)
        for k in keys:
            p = self._cfg.get(k)
            v = revids.get(k, p)
            if p != v:
                result = True
            new_cfg[k] = v
        if result:
            self._cfg = new_cfg
        return result

##############################################################################

CONFIG_TMPL = """\
# See the documentation for details on the project.cfg format
# For any component you should specify at least RELPATH and BRANCH_URL option

## describe each component in subsequent sections, e.g.:
# [component "name"]
# RELPATH =
# BRANCH_URL =

"""

LOCAL_TMPL = """\
# local.cfg v1

[URL]
PROJECT =
SNAPSHOT =

[CURRENT]
ALT =
SUBSET =

[CUSTOM]

[LOCAL BRANCHES]

[SETTINGS]
"""
