# Simple support for organizing vcs branches as complex project
#
# Copyright (C) 2008-2011 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
import os
import shutil
import tempfile

from bzrlib import (
    branch,
    bzrdir,
    errors,
    ignores,
    osutils,
    trace,
    transport,
    urlutils,
    workingtree,
    )
from bzrlib.commands import run_bzr
from bzrlib.transport import local as _mod_local

from bzrlib.plugins import scmproj
from bzrlib.plugins.scmproj import (
    action,
    config,
    vcs,
    version_info,
    )
from bzrlib.plugins.scmproj.action import ActionList, url_for_display
from bzrlib.plugins.scmproj.vcs import (
    BzrVcs,
    bzr_vcs,
    checkers_reg,
    vcs_reg,
    )
""")

# Project v1

BASE       = '.scmproj'
CONFIG     = 'project.cfg'
ALTS_DIR   = 'alts'
ALTS_INDEX = 'index'

CONFIG_FN  = BASE + '/' + CONFIG    # candidate for removal

README_FILE = 'readme.txt'
README_TEXT = """\
This is a control directory for a project.

You need to have the bzr scmproj plugin installed
to be able work with the project itself.

See http://launchpad.net/bzr-scmproj
for more information about the scmproj plugin.
"""

# Project v2

BZRMETADIR = '.bzrmeta'
SCMPROJDIR = 'scmproj'
SNAPSHOT_FILE = 'snapshot'
CONFIG_V2 = BZRMETADIR + '/' + SCMPROJDIR + '/' + CONFIG
SNAPSHOT_V2 = BZRMETADIR + '/' + SCMPROJDIR + '/' + SNAPSHOT_FILE


class SilentStream(object):
    def write(self, s):
        pass
silent_file = SilentStream()


def is_url_local(url):
    return isinstance(transport.get_transport(url),
        _mod_local.LocalTransport)


def safe_path(name):
    """Encode name as safe cross-platform path for filesystem."""
    # this function implemented looking at hg store.py code
    # but we are using unicode instead of utf-8
    # so we can luxury to write it a bit shorter
    res = unicode(name).strip().translate(_SAFE_MAP)    # see end of the file
    parts = res.split('.', 1)
    prefix = parts[0]
    # special windows names (like aux or nul) should be mangled
    if prefix in _WINDOWS_RESERVED_FILENAMES:           # see end of the file
        encoded = prefix[:-1] + '~%02x' % ord(prefix[-1])
        parts[0] = encoded
        res = '.'.join(parts)
    if res[-1] == '.':
        res = res[:-1] + '~2e'     # 2e is hex code for dot
    return res


class NullConfig(object):
    """Special config to catch and report programming errors
    when we try to work with the project witout opened it first.
    It will raise BzrError on any access.
    """
    def __setattr__(self, k, v):
        raise errors.BzrError('Internal error: '
            'Project should be opened first')
    def __getattribute__(self, k):
        raise errors.BzrError('Internal error: '
            'Project should be opened first')


class CutLine(object):
    """Cut line helper.
    Print cut line to output stream starting from second call.
    """
    def __init__(self, stream, width=60):
        self.stream = stream
        self.line = '#%s\n' % ('-'*width)
        self.emit = False

    def __call__(self, force=False):
        if self.emit or force:
            self.stream.write(self.line)
        else:
            self.emit = True


class ProjectCommand(object):
    """Abstract high-level command to execute on project
    and all its subprojects.
    Make sure you pass down _cmd arguments (which is equal to self)
    so every subproject can execute this command recursively.
    """

    def __init__(self, **kw):
        if '_cmd' in kw:
            raise errors.BzrInternalError('_cmd in the kw dict')
        self.kw = {}
        self.kw.update(kw)
        self.kw['_cmd'] = self
        self.clone_subproject = False  # used in pget to force cloning subprojects

    def execute(self, po):
        """Execute this command for Project object po.
        Apply the arguments from self.kw.
        Override this method in your subclasses.
        """
        raise NotImplementedError


class RunActionsCommand(ProjectCommand):
    """Default command to run_actions in every subproject."""

    def __init__(self, action_list, **kw):
        super(RunActionsCommand, self).__init__(**kw)
        self.action_list = action_list

    def execute(self, po):
        return po.run_actions(self.action_list, **self.kw)


class ProjectNotFound(errors.BzrError):
    _fmt = "Project not found at %(url)s"


class Project(object):
    """Main class which knows how to work with projects.

    Transparently supports:
    * v1 projects with project config in separate control branch (.scmproj),
      both local and remote variants
    * v2 projects with project config directly in the top-level branch
      (file .bzrmeta/scmproj/project.cfg)
    """

    def __init__(self, url):
        """@param url: project root url."""
        self.branch = None
        self.cfg = NullConfig()
        self.settings = NullConfig()
        self.t = transport.get_transport(url)
        self.i_am_subproject = False
        self._is_local = isinstance(self.t, _mod_local.LocalTransport)
        self._format = None
        self.index_card = config.IndexCard()

    format = property(fget=lambda self: self._format, doc="Project format")

    def open(self):
        """Open the project for business."""
        found, format, opened_project = self.find_project(need_open=True,
            raise_error=True)
        self._format = format
        return opened_project

    def clone(self, to_url, revision_id=None, revision=None):
        """Clone project to local copy."""
        if revision_id is not None and revision is not None:
            raise errors.BzrError('You should specify either revision_id '
                'or revision.')
        found, format, _unused = self.find_project(raise_error=True)
        if format == 'v2':
            self._clone_v2(to_url, revision_id=revision_id, revision=revision)
        elif format == 'v1':
            if revision_id is not None or revision is not None:
                raise errors.BzrError('Projects v1 does not support clone '
                    'of specified revision.')
            if self._is_local:
                return self._clone_v1_local(to_url)
            else:
                return self._clone_v1_remote(to_url)
        else:
            raise errors.BzrError('Unknown project format: %s' % format)

    @staticmethod
    def _create_skeleton_v2(path, branch_format=None, need_commit=False):
        """Create .scmproj branch locally.
        @param path: either path or URL.
        """
        t = transport.get_transport(path)
        t.ensure_base()
        # create branch
        vo = BzrVcs()
        try:
            vo.create_branch(path, branch_format)
        except errors.BranchExistsWithoutWorkingTree:
            raise errors.BzrError('Branch already exists but has no working tree at: %s' % path)
        except errors.AlreadyBranchError:  # that's ok, see bug #557089
            # check if it's already initialized
            is_project = Project(t.base).find_project(need_open=False, raise_error=False)[0]
            if is_project:
                raise errors.BzrError('Project already initialized at: %s' % path)
        # create mandatory files
        wt = workingtree.WorkingTree.open(path)
        if not t.has(BZRMETADIR):
            t.mkdir(BZRMETADIR)
        scmproj_dir = osutils.pathjoin(BZRMETADIR, SCMPROJDIR)
        if not t.has(scmproj_dir):
            t.mkdir(scmproj_dir)
        t.put_bytes(CONFIG_V2, config.Config().to_str())
        wt.add([BZRMETADIR, scmproj_dir, CONFIG_V2])
        # add empty snapshot file from the beginning, bug #563941
        root_path = t.local_abspath('.')
        sn = SnapshotV2(root_path)
        sn.set_index_card(config.IndexCard())   # snapshot file automatically added to wt
        if need_commit:
            wt.commit('initialization')

    @staticmethod
    def _create_skeleton_v1(path, branch_format=None, _unused_need_commit=None):
        """Create .scmproj branch locally."""
        t = transport.get_transport(path)
        t.ensure_base()
        # create branch
        vo = BzrVcs()
        root = osutils.pathjoin(path, BASE)
        try:
            vo.create_branch(root, branch_format)
        except (errors.AlreadyBranchError,
                errors.BranchExistsWithoutWorkingTree):
            raise errors.BzrError(BASE+' directory already intialized.')
        # create mandatory files
        t = transport.get_transport(root)
        wt = workingtree.WorkingTree.open(root)
        t.put_bytes(CONFIG, config.Config().to_str())
        wt.add([CONFIG])
        t.mkdir(ALTS_DIR)
        wt.add([ALTS_DIR])
        index = ALTS_DIR + '/' + ALTS_INDEX
        t.put_bytes(index, '')
        wt.add([index])
        t.put_bytes(README_FILE, README_TEXT)
        wt.add([README_FILE])
        wt.commit('initialization')

    @staticmethod
    def initialize(url, branch_format=None, _scmproj_format='v2',
        _need_commit=False):
        """Initialize new scmproj control dir."""
        _create_skeleton = getattr(Project,
            '_create_skeleton_%s' % _scmproj_format, None)
        if _create_skeleton is None:
            raise errors.BzrError('Wrong project format: %s' % _scmproj_format)
        t = transport.get_transport(url)
        if isinstance(t, _mod_local.LocalTransport):
            _create_skeleton(t.abspath('.'), branch_format, _need_commit)
        else:
            try:
                branch.Branch.open(url)
            except errors.NotBranchError:
                pass
            else:
                raise errors.BzrError('A branch already present at %s' % url)
            # create control dir in the temp dir and then push it to server
            temp_dir = tempfile.mkdtemp()
            _create_skeleton(temp_dir, branch_format, True)
            from_url = urlutils.join(urlutils.local_path_to_url(temp_dir), BASE)
            BzrVcs().push(url, from_url, use_existing_dir=True)
            try:
                shutil.rmtree(temp_dir)
            except (IOError, OSError), e:
                trace.mutter('Unable to delete temp directory %r: error: %s',
                    temp_dir, str(e))

    @staticmethod
    def open_containing(url):
        """Try to open project walking upwards in directory tree.
        @return: project opened for business
        """
        if not is_url_local(url):
            raise errors.BzrError("Remote project cannot be opened.")
        t = transport.get_transport(url)
        path = urlutils.local_path_from_url(t.base)
        while True:
            try:
                return Project(path).open()
            except ProjectNotFound:
                new_path = osutils.abspath(osutils.pathjoin(path, '..'))
                if new_path == path:
                    raise errors.BzrError('Not a project: "%s".' % url)
                path = new_path

    def save_config(self):
        self.t.put_bytes(CONFIG_V2, self.cfg.to_str())

    def save_settings(self, alt=None, subset=None,
        override=False):
        """Save local settings values in setings config (branch.conf)."""
        if self._format == 'v2':
            return
        if alt:
            if not self.settings.alt or override:
                self.settings.alt = alt
        if subset:
            if not self.settings.subset or override:
                self.settings.subset = subset

    def get_alt_name(self):
        """Return current alt name."""
        return self.settings.alt or self.cfg.get_alt_default_link()

    def get_subset_name(self):
        """Return current subset name."""
        return self.settings.subset or self.cfg.get_subset_default_link()

    def info(self, to_file=None, indent=None):
        if to_file is None:
            import sys
            to_file = sys.stdout
        if indent is None:
            indent = ''

        cfg = self.cfg
        settings = self.settings
        url = self.t.base

        to_file.write(indent+'Project: ')
        if cfg.name:
            to_file.write(cfg.name)
        else:
            to_file.write('unnamed')
        to_file.write('\n')
        to_file.write(indent+'URL: '+urlutils.unescape_for_display(url, 'utf-8')+'\n')
        if cfg.description:
            to_file.write(indent+'Description: '+cfg.description+'\n')
        if settings.project_url not in ('.', '', None):
            to_file.write(indent+'Parent URL: '+urlutils.unescape_for_display(
                settings.project_url, 'utf-8')+'\n')
        if settings.alt or settings.subset:
            to_file.write('\n')
        if settings.alt:
            to_file.write(indent+'Alt: %s\n' % settings.alt)
        if settings.subset:
            to_file.write(indent+'Subset: %s\n' % settings.subset)

        alt_dict = cfg.get_alt(self.get_alt_name())

        if cfg.components:
            to_file.write('\n'+indent+'Components:\n')
            for kname in self.get_component_list():
                to_file.write(indent+'  [%s]\n' % kname)
                bname = alt_dict[kname]
                if bname is None:       # this component is missed in the alt
                    to_file.write(indent+'      (missing)\n')
                    continue
                ko = Component(cfg, kname)
                kb = ko.get_branch(bname)
                to_file.write(indent+'      relpath = %s\n' % ko.relpath)
                to_file.write(indent+'      branch  = %s\n' % kb.url)

        project_root = urlutils.local_path_from_url(url)

        if cfg.subprojects:
            to_file.write('\n'+indent+'Subprojects:\n')
            for sname in self.get_subprojects_list():
                to_file.write(indent+'  [%s]\n' % sname)
                subp = Subproject(cfg, sname, project_root).get_instance()
                if subp is None:
                    to_file.write(indent+'      (missing)\n')
                    continue
                subp.info(to_file, indent+' '*6)

    def run_cmds(self, command_list, to_file=None, ignore_errors=False,
        alt_name=None, components=None, cd_relpath=False, verbose=True,
        dry_run=False, shell=False, vars_dict=None, checker_func=None,
        _cmd=None):
        if self._format == 'v2' and alt_name is not None:
            raise errors.BzrError('Project v2 does not support alts.')
        if not shell:
            action_class = action.BzrCommandsAction
        else:
            action_class = action.ShellCommandsAction
        action_list = ActionList([action_class(command_list,
            ignore_errors, verbose, dry_run, vars_dict)])
        return self.run_actions(action_list, to_file, alt_name,
            components, cd_relpath=cd_relpath, verbose=verbose,
            checker_func=checker_func, _cmd=_cmd)

    def run_actions(self, action_list, to_file=None, alt_name=None,
        component_list=None, cd_relpath=False, verbose=True,
        checker_func=None, _cmd=None):
        """Generic iterator over components to execute some actions.

        @param  action_list:    list of Action instances to execute.
        @param  to_file:        stream to write messages.
        @param  component_list: optional list of components to iterate.
        @param  alt_name:       name of the alt to select component branches.
        @param  cd_relpath:     execute action in component working copy
                                (cd {RELPATH} before execution).
        @param  verbose:        show verbose info about actions
        @param  checker_func:   function to check component before execute
                                action(s); if function returns True then
                                component will be processed.

        @param _cmd:            ProjectCommand object to execute
                                high-level commands for each subproject.

        @return: None if actions executed successfully for all components,
            otherwise return code will be used as return code for CLI.
        """
        if self._format == 'v2' and alt_name is not None:
            raise errors.BzrError('Project v2 does not support alts.')
        if to_file is None:
            to_file = silent_file
        cutline = CutLine(to_file)
        if alt_name is None:
            alt_name = self.get_alt_name()
        alt_dict = self.cfg.get_alt(alt_name)
        if component_list is None:
            component_list = self.cfg.get_full_set()
        else:
            self.cfg.ensure_components(component_list)
        if checker_func is None:
            checker_func = checkers_reg.get('all')

        if _cmd is None:
            _cmd = RunActionsCommand(action_list,
                to_file=to_file, alt_name=None, component_list=None,
                cd_relpath=cd_relpath, verbose=verbose,
                checker_func=checker_func)

        project_root = urlutils.local_path_from_url(self.t.base)

        # local_paths dictionary
        # there should be paths in context of current component/kbranch
        local_paths = {
            'PROJECT_ROOT': project_root,
            }
        local_urls = {
            'PROJECT_ROOT': self.t.base,
            }
        orig_cwd = os.getcwdu()
        os.chdir(project_root)
        # execute prolog
        for func in getattr(action_list, 'prolog', []):
            cutline()
            func()
        # let's go
        # components first
        for kname in component_list:
            trace.mutter('Component: %s', kname)

            ko = Component(self.cfg, kname, self.index_card)

            local_tree = osutils.pathjoin(project_root, ko.relpath)
            tree_url = urlutils.local_path_to_url(local_tree)
            local_paths['LOCAL_TREE'] = local_tree
            local_urls['LOCAL_TREE'] = tree_url
            local_paths['LOCAL_BRANCH'] = local_tree
            local_urls['LOCAL_BRANCH'] = tree_url

            bname = alt_dict[kname]
            if bname is None:       # 
                trace.mutter('this component is missed in the alt')
                continue

            kb = ko.get_branch(bname)

            if not checker_func(kb, local_tree, kb.url):
                trace.mutter('component skipped')
                continue

            cutline()
            to_file.write('# Component: %s\n' % kname)

            if cd_relpath and ko.relpath != '.':
                if verbose:
                    to_file.write("#  Entering directory `%s'\n" % ko.relpath)
                try:
                    os.chdir(local_tree)
                except OSError:
                    # can't change dir because there is no component tree
                    to_file.write("#  Component `%s' not found.\n" % (kname,))
                    continue
            try:
                # we ready to run actions for the component branch
                for action in action_list:
                    ret_code = action(to_file, ko, kb, local_paths, local_urls)
                    if ret_code is not None:
                        return ret_code
            finally:
                if cd_relpath:
                    os.chdir(orig_cwd)
        # subprojects in second pass
        for sname in self.get_subprojects_list():
            cutline()
            to_file.write('# Subproject: %s\n' % sname)
            sub_object = Subproject(self.cfg, sname, project_root,
                root_revision_id=self.index_card.get_revid(sname))
            subp = sub_object.get_instance()
            if subp is None:
                if not _cmd.clone_subproject:
                    if verbose:
                        to_file.write('#  Subproject is missing. Operation skipped\n')
                    continue
                else:
                    subp = sub_object.clone_and_open()
            _cmd.rev_str = self.index_card.get_revision(sname)
            if verbose:
                to_file.write("#  Entering directory `%s'\n" %
                    self.cfg.get_subproject_relpath(sname))
            subp.chdir_to_root()
            try:
                cutline()
                ret_code = _cmd.execute(subp)
                if ret_code is not None:
                    return ret_code
            finally:
                os.chdir(orig_cwd)
        # execute epilog
        os.chdir(project_root)
        try:
            for func in getattr(action_list, 'epilog', []):
                cutline()
                func()
        finally:
            os.chdir(orig_cwd)
        if not self.i_am_subproject:
            cutline()
            to_file.write('# Done.\n')

    def run_cmds_v2(self, command_list, to_file=None, ignore_errors=False,
        includes=None, excludes=None, cd_relpath=False, verbose=True,
        dry_run=False, shell=False, vars_dict=None, checker_func=None):
        if not shell:
            action_class = action.BzrCommandsActionV2
        else:
            action_class = action.ShellCommandsActionV2
        action_list = ActionList([action_class(command_list,
            ignore_errors, verbose, dry_run, vars_dict)])
        ptree = ProjectTree(self, True)
        ptree.build_tree()
        return self.run_actions_v2(action_list, 
            ptree.iterate_filter(includes=includes, excludes=excludes),
            to_file=to_file, cd_relpath=cd_relpath, verbose=verbose,
            checker_func=checker_func)

    def run_actions_v2(self, action_list, iter_nodes,
        to_file=None, cd_relpath=False, verbose=True,
        checker_func=None):
        """Execute actions for ProjectNodes.

        @param  action_list:    list of Action instances to execute.
        @param  iter_nodes:     iterator from ProjectTree, emits ProjectNodes.

        @param  to_file:        stream to write messages.
        @param  cd_relpath:     execute action in component working copy
                                (cd {RELPATH} before execution).
        @param  verbose:        show verbose info about actions
        @param  checker_func:   function to check component before execute
                                action(s); if function returns True then
                                component will be processed.

        @return: None if actions executed successfully for all components,
            otherwise return code will be used as return code for CLI.
        """
        if to_file is None:
            to_file = silent_file
        cutline = CutLine(to_file)

        if checker_func is None:
            checker_func = checkers_reg.get('all')

        project_root_path = urlutils.local_path_from_url(self.t.base)
        project_root_url = self.t.base

        orig_cwd = os.getcwdu()
        os.chdir(project_root_path)
        # execute prolog
        for func in getattr(action_list, 'prolog', []):
            cutline()
            func()
        # let's go
        for node in iter_nodes:
            trace.mutter('Component: %s', node.name)
            local_tree = node.abspath
            tree_url = urlutils.local_path_to_url(local_tree)
            local_paths = {
                'PROJECT_ROOT': project_root_path,
                'LOCAL_TREE': local_tree,
                'LOCAL_BRANCH': local_tree,
                }
            local_urls = {
                'PROJECT_ROOT': project_root_url,
                'LOCAL_TREE': tree_url,
                'LOCAL_BRANCH': tree_url,
                }

            if node.is_subproject and action_list.subproject:
                node_actions = action_list.subproject
            else:
                kb = node.kbranch
                if kb is None:
                    trace.mutter('this component has no kbranch')
                    continue
    
                if not checker_func(kb, local_tree, kb.url):
                    trace.mutter('component skipped')
                    continue

                node_actions = action_list

                cutline()
                to_file.write('# Component: %s\n' % node.name)
    
            if cd_relpath and node.relpath != '.':
                if verbose:
                    to_file.write("#  Entering directory `%s'\n" % node.relpath)
                try:
                    os.chdir(local_tree)
                except OSError:
                    # can't change dir because there is no component tree
                    to_file.write("#  Component `%s' not found.\n" % (node.name,))
                    continue
            try:
                # we ready to run actions for the component branch
                for action in node_actions:
                    ret_code = action(to_file, node, local_paths, local_urls)
                    if ret_code is not None:
                        return ret_code
            finally:
                if cd_relpath:
                    os.chdir(orig_cwd)
            if node.is_subproject:
                cutline()
        # execute epilog
        os.chdir(project_root_path)
        local_paths = {
            'PROJECT_ROOT': project_root_path,
            }
        local_urls = {
            'PROJECT_ROOT': project_root_url,
            }
        try:
            for func in getattr(action_list, 'epilog', []):
                cutline()
                func(to_file, local_paths, local_urls)
        finally:
            os.chdir(orig_cwd)
        if not self.i_am_subproject:
            cutline()
            to_file.write('# Done.\n')

    def update_workspace(self, to_file=None, alt_name=None,
        component_list=None, all=False, _cmd=None):
        """Update local project copy: create and update component trees."""
        if self._format == 'v2' and alt_name is not None:
            raise errors.BzrError('Project v2 does not support alts.')
        action_list = ActionList([
            action.UpdateWorkingTreeAction(create=True),
            ])
        checker_func = checkers_reg.get('different-tips')
        if all:
            checker_func = checkers_reg.get('all')
        return self.run_actions(action_list, to_file, alt_name, component_list,
            checker_func=checker_func, _cmd=_cmd)

    def update_control_branch(self, to_file=None, rev_str=None):
        """Special pre-action to update control branch with config.
        Should be executed before any other project actions
        because it may re-read config.
        @param rev_str: revision as string to pass it down to vcs layer.
        """
        if to_file is None:
            to_file = silent_file
        from_url = self.branch.get_parent()
        if from_url:
            to_file.write('# Updating control branch from %s...\n' % (
                action.url_for_display(from_url)))
            base_url = self.t.base
            if self._format == 'v1':
                base_url = urlutils.join(self.t.base, BASE)
            bzr_vcs.pull(base_url, from_url, revision=rev_str)
            # project config could be changed by this action, need to re-read it
            self.open()
        else:
            to_file.write('# Skip control branch: no parent location known.\n')
        CutLine(to_file)(force=True)

    def commit(self, msgfile, to_file=None, component_list=None,
        components_only=False, _cmd=None):
        """Commit the project: changes in components and control branch."""
        action_list = ActionList([
            action.CommitAction(osutils.abspath(msgfile)),
            ])
        if not components_only:
            action_list.epilog = [
                self._get_commit_control_branch_action(msgfile, to_file)]
        return self.run_actions(action_list, to_file,
            component_list=component_list,
            checker_func=vcs.check_changed_wo_unknowns, _cmd=_cmd)

    def _get_commit_control_branch_action(self, msgfile, to_file):
        def commit_control_branch():
            vo = BzrVcs()
            path = urlutils.local_path_from_url(self.branch.base)
            if vo.is_tree_changed(path):
                vo.commit_tree(path, msgfile)
        return commit_control_branch

    def commit_v2(self, msgfile, to_file=None, component_list=None):
        ptree = ProjectTree(self, toplevel=True)
        ptree.build_tree()
        checklist = ProjectTree.build_checklist(component_list)
        self.commit_project_tree_v2(ptree, msgfile, to_file, checklist)

    def commit_project_tree_v2(self, project_tree, msgfile, to_file=None,
        checklist=None):
        commit_action = action.CommitAndCollectRevidActionV2(msgfile,
            seed_revids=project_tree.collect_revids())
        epilog_action = action.UpdateAndCommitSnapshotEpilogV2(commit_action)
        action_list = ActionList(
            sequence=commit_action,
            epilog=epilog_action,
            subproject=commit_action,
            )
        iterator = project_tree.iterate_filter_chk_reverse(
            yield_subproject_node=True, recurse=False,
            includechecklist=checklist)
        ret_code = self.run_actions_v2(action_list, iterator, to_file,
            checker_func=vcs.check_changed_wo_unknowns,
            )
        if ret_code not in (None, 0):
            return None

    def get_component_list(self, subset=None, full_set=None,
        includes=None, excludes=None):
        """Return component list based on subset.
        If neither subset nor full_set specified then components of current
        project subset is returned.
        """
        if self._format == 'v2':
            if subset is not None or full_set is not None:
                raise errors.BzrError('Project v2 does not support neither '
                    'subset nor full-set options.')
        if full_set:
            subset = 'FULL'
        elif subset is None:
            subset = self.get_subset_name()
        component_list = self.cfg.get_subset(subset)
        if includes or excludes:
            r = set(component_list)
            all = set(self.cfg.components)
            if includes:
                for i in includes:
                    if i in all:
                        r.add(i)
                    else:
                        raise errors.BzrError('Unknown component: '+i)
            if excludes:
                for e in excludes:
                    if e in all:
                        if e in r:
                            r.remove(e)
                    else:
                        raise errors.BzrError('Unknown component: '+e)
            component_list = self.cfg.sort_components(r)
        return component_list

    def get_subprojects_list(self):
        return self.cfg.sort_subprojects()

    def cli_helper(self, alt=None, subset=None, full_set=None,
        includes=None, excludes=None):
        """@return: (alt, component_list)."""
        if self._format == 'v2':
            if alt is not None or subset is not None or full_set is not None:
                raise errors.BzrError('Project v2 does not support neither '
                    'alt nor subset nor full-set options.')
            return (None, self.get_component_list())
        if alt is None:
            alt = self.get_alt_name()
        component_list = self.get_component_list(subset, full_set,
            includes, excludes)
        return (alt, component_list)

    def vars_helper(self, **kw):
        """@return: dict with additional template variables."""
        d = {}
        for k in kw:
            if kw[k] is None:
                continue
            d[k] = kw[k]            # {varN}
            d[k.upper()] = kw[k]    # {VARN}  -- only for projects v1
        return d

    def publish(self, to_file=None, alt_name=None,
        component_list=None, components_only=False, strict=None, all=False,
        _cmd=None):
        """Push changes from local component branches to their origins."""
        if self._format == 'v2' and alt_name is not None:
            raise errors.BzrError('Project v2 does not support alts.')
        action_list = ActionList([action.PublishKBranch(strict)])
        if not components_only:
            action_list.epilog = [
                self._get_publish_control_branch_action(to_file, strict=strict)]
        checker_func = vcs.check_different_tips_or_changes
        if all:
            checker_func = checkers_reg.get('all')
        return self.run_actions(action_list, to_file, alt_name, component_list,
            checker_func=checker_func, _cmd=_cmd)

    def _get_publish_control_branch_action(self, to_file=None, strict=None):
        def publish_control_branch():
            #to_url = self.settings.project_url
            to_url = self.branch.get_push_location() or self.settings.project_url
            if to_url:
                to_file.write('# Push control branch to %s...\n' % (
                    action.url_for_display(to_url)))
                vo = BzrVcs()
                vo.push(to_url, self.branch.base, strict=strict)
            else:
                to_file.write('# Skip control branch: no push location known.\n')
        return publish_control_branch

    @staticmethod
    def _clone_v1_to_local_url(br_from, revision_id, to_url, accelerator_tree=None):
        """Internal machinery of clone (local,remote) -> local.
        @param  br_from:    source branch instance
        @param  revision_id: tip of .scmproj branch to clone.
        @param  to_url:     destination local URL
        @param  accelerator_tree:   can be used for local->local clone.
        """
        br_from.lock_read()
        try:
            # create base dir if needed
            transport.get_transport(to_url).ensure_base()
            to_ctrl_url = urlutils.join(to_url, BASE)
            try:
                transport.get_transport(to_ctrl_url).mkdir('.')
            except errors.FileExists:
                raise errors.BzrError('Target directory "%s" already'
                    ' exists.' % url_for_display(to_ctrl_url))
            br_from.bzrdir.sprout(to_ctrl_url,
                revision_id=revision_id,
                force_new_repo=True,
                accelerator_tree=accelerator_tree, source_branch=br_from)
        finally:
            br_from.unlock()

    @staticmethod
    def _clone_v2_to_local_url(br_from, revision_id, to_url, accelerator_tree=None):
        """Internal machinery of clone (local,remote) -> local.
        @param  br_from:    source branch instance
        @param  revision_id: tip of .scmproj branch to clone.
        @param  to_url:     destination local URL
        @param  accelerator_tree:   can be used for local->local clone.
        """
        br_from.lock_read()
        try:
            # create base dir if needed
            transport.get_transport(to_url).ensure_base()
            br_from.bzrdir.sprout(to_url,
                revision_id=revision_id,
                #force_new_repo=True,     # using shared repo if possible
                accelerator_tree=accelerator_tree, source_branch=br_from)
        finally:
            br_from.unlock()

    def chdir_to_root(self):
        project_root = urlutils.local_path_from_url(self.t.base)
        os.chdir(project_root)

    def _clone_v1_local(self, to_url):
        """clone local->local."""
        from_url = urlutils.join(self.t.base, BASE)
        accelerator_tree, br_from = bzrdir.BzrDir.open_tree_or_branch(from_url)
        revision_id = br_from.last_revision()
        Project._clone_v1_to_local_url(br_from, revision_id, to_url, accelerator_tree)

    def _open_v1_local(self, _branch=None):
        """Open project for business.
        @return: self.
        """
        if _branch is None:
            self.branch = branch.Branch.open(self.t.abspath(BASE))
        else:
            self.branch = _branch
        # read and parse project.cfg
        self.cfg = config.Config(self.t.get_bytes(CONFIG_FN))
        self.settings = config.LocalSettings(self.branch)
        return self

    def _clone_v1_remote(self, to_url):
        """clone remote->local."""
        from_url = self.t.base
        br_from = branch.Branch.open(from_url)
        revision_id = br_from.last_revision()
        Project._clone_v1_to_local_url(br_from, revision_id, to_url)

    def _open_v1_remote(self, _branch=None):
        raise errors.BzrError("Remote project cannot be opened.")

    def _open_v2_local(self, _branch):
        """Open project for business.
        @return: self.
        """
        self.branch = _branch
        # read and parse project.cfg
        self.cfg = config.Config(self.t.get_bytes(CONFIG_V2))
        self.settings = config.LocalSettingsV2(self.branch)
        snapshot = SnapshotV2(urlutils.local_path_from_url(self.t.base))
        self.index_card = snapshot.get_index_card()
        return self

    def _clone_v2(self, to_url, revision_id=None, revision=None):
        from_url = self.t.base
        accelerator_tree, br_from = bzrdir.BzrDir.open_tree_or_branch(from_url)
        if revision_id is None:
            if revision is not None:
                revision_id = revision.as_revision_id(br_from)
            else:
                revision_id = br_from.last_revision()
        Project._clone_v2_to_local_url(br_from, revision_id, to_url, accelerator_tree)

    def find_project(self, need_open=False, raise_error=False):
        """Find project in the location specified in the Project constructor.
        @param need_open: if True then found project returned as 3rd element
            of the result.
        @param raise_error: enable raising error ProjectNotFound if project
            not found.
        @return: (found[True|False], format(v2|v1|None), opened project|None)
        NOTE: remote projects cannot be opened, so if need_open is True while
        opening remote project you'll get an exception.
        """
        found = False
        format = None
        opened_project = None

        base_url = self.t.base
        # v2
        if self._is_local:
            try:
                tree, br = bzrdir.BzrDir.open_tree_or_branch(base_url)
            except errors.NotBranchError:
                tree, br = None, None
        else:
            tree = None
            try:
                br = branch.Branch.open(base_url)
            except errors.NotBranchError:
                br = None
        if tree:
            tree.lock_read()
            try:
                if tree.path2id(CONFIG_V2) is not None:
                    found = True
                    format = 'v2'
            finally:
                tree.unlock()
        elif br:
            br.lock_read()
            try:
                rt = br.repository.revision_tree(br.last_revision())
                if rt.path2id(CONFIG_V2) is not None:
                    found = True
                    format = 'v2'
            finally:
                br.unlock()
        if found:
            if need_open:
                if tree:
                    opened_project = self._open_v2_local(br)
                else:
                    raise errors.BzrError(
                        "Project without working tree cannot be opened.")
            return (found, format, opened_project)

        # v1
        if self._is_local:
            base_url = urlutils.join(self.t.base, BASE)
        try:
            br = branch.Branch.open(base_url)
        except errors.NotBranchError:
            pass
        else:
            br.lock_read()
            try:
                rt = br.repository.revision_tree(br.last_revision())
                if rt.path2id(CONFIG) is not None:
                    found = True
                    format = 'v1'
            finally:
                br.unlock()
            if found and need_open:
                if self._is_local:
                    opened_project = self._open_v1_local(br)
                else:
                    opened_project = self._open_v1_remote(br)
        if not found and raise_error:
            raise ProjectNotFound(url=self.t.base)
        return (found, format, opened_project)

    def ignore_relpath(self, relpath):
        wt = workingtree.WorkingTree.open(self.t.base)
        ignores.tree_ignores_add_patterns(wt, ['./'+relpath])


class Component(object):
    """Project component."""

    def __init__(self, config_object, name, index_card=None):
        """Create component object based on data from project.cfg.
        @param  config_object:  Config (project.cfg wrapper) instance
        @param  name:   name of the component
        """
        self.name = name
        self._cfg = config_object
        self._data = config_object.get_component_section(name)
        self.description = self._data.get('DESCRIPTION')
        self.vcs = self._data.get('VCS', config_object.get_default_vcs())
        if self.vcs not in ('bzr', None):
            raise errors.BzrError('Unsupported VCS: %s' % self.vcs)
        self.format = (self._data.get('FORMAT')
            or config_object.get_default_format())
        self.branches = self._cfg.get_component_branches(self.name)
        if index_card is not None:
            self._index_card = index_card
        else:
            self._index_card = config.IndexCard()

    def _get_relpath(self):
        return self._cfg.get_component_relpath(self.name)

    relpath = property(_get_relpath)

    def get_branch(self, branch_name=None):
        """@return: KBranch object"""
        if branch_name in ('DEFAULT', None):
            branch_name = self.branches[0]
        if branch_name in self.branches:
            sb = self._cfg.get_branch_section(self.name, branch_name)
            revision = self._index_card.get_revision(self.name)
            if sb is None:
                # simple branch
                return KBranch(self._cfg, self.name, branch_name,
                    self.vcs, revision=revision, format=self.format)
            else:
                # explicitly defined branch
                vcs = sb.get('VCS', self.vcs)
                format = sb.get('FORMAT', self.format)
                revision = sb.get('REVISION', revision)
                extra = sb
                return KBranch(self._cfg, self.name, branch_name,
                    vcs, revision=revision,
                    format=format, extra=extra)
        raise KeyError


class KBranch(object):
    """Component branch."""

    is_root = False

    def __init__(self, cfg, component_name, name,
            vcs, revision=None, format=None, extra=None):
        self._cfg = cfg
        self.component_name = component_name
        self.name = name
        self.vcs = vcs
        if self.vcs not in ('bzr', None):
            raise errors.BzrError('Unsupported VCS: %s' % self.vcs)
        self.format = format
        self.revision = revision
        self.extra = extra or {}

    def _get_url(self):
        return self._cfg.get_component_branch_url(self.component_name,
            self.name)

    url = property(_get_url)


class RootComponent(object):
    def __init__(self, proj, name=None, revision=None):
        self._project = proj
        if name is None:
            name = 'root'
        self.name = name
        self.description = None
        self.vcs = 'bzr'
        self.format = None
        self.branches = ['DEFAULT']
        self.relpath = '.'
        if self._project.format == 'v1':
            self.relpath = BASE
        self._revision = revision

    def get_branch(self):
        return RootKBranch(self._project, self.name, revision=self._revision)

    def is_v1(self):
        return self._project.format == 'v1'


class RootKBranch(object):

    is_root = True

    def __init__(self, proj, component_name, url=None, revision=None):
        self._project = proj
        self.component_name = component_name
        if self._project.format == 'v2':
            self.name = 'root branch'
        else:
            self.name = 'control branch'
        self.vcs = 'bzr'
        self.format = None
        self.revision = revision
        self.extra = {}
        self._url = url

    def _get_url(self):
        return self._url or self._project.settings.project_url

    url = property(_get_url)


class Subproject(object):
    """Helper to get subproject's Project instance based on main config."""

    def __init__(self, cfg, subproject_name, super_root, root_revision_id=None):
        """Create subproject helper from super-project data.
        @param cfg: super-project config.
        @param subproject_name: name of subproject in the config.
        @param super_root: root directory of super-project.
        """
        self._cfg = cfg
        self.name = subproject_name
        self.super_root = super_root
        self.root_revision_id = root_revision_id

    def __repr__(self):
        r = ['Subproject(']
        r.append('name=')
        r.append(self.name)
        if self.root_revision_id:
            r.append(',revid=')
            r.append(self.root_revision_id)
        r.append(')')
        return ''.join(r)

    def _get_relpath(self):
        return self._cfg.get_subproject_relpath(self.name)

    relpath = property(_get_relpath)

    def _get_sub_root(self):
        return osutils.pathjoin(self.super_root, self.relpath)

    sub_root = property(_get_sub_root)

    def _get_url(self):
        return self._cfg.get_subproject_url(self.name)

    url = property(_get_url)

    def get_instance(self, auto_clone=False):
        """@return:  Project object instance of this subproject."""
        try:
            subp = Project(self.sub_root).open()
            subp.i_am_subproject = True
            return subp
        except ProjectNotFound:
            # there is no subproject control branch,
            # so we have to clone it first if auto_clone enabled
            if not auto_clone:
                return None
            return self.clone_and_open()

    def clone(self):
        """Clone subproject to directory relative to super project root."""
        from_url = self.url
        to_path = self.sub_root
        if not to_path.startswith(self.super_root) or to_path == self.super_root:
            raise errors.BzrError(
                "Subproject %r clone error: wrong relpath: %r",
                self.name, self.relpath)
        Project(from_url).clone(to_path, revision_id=self.root_revision_id)

    def clone_and_open(self):
        self.clone()
        subp = Project(self.sub_root).open()
        subp.i_am_subproject = True
        return subp


class ProjectNode(object):
    __slots__ = ['component', 'kbranch', 'root_abspath', 'prefix',
        'is_root', 'is_toplevel', 'is_subproject',
        'subproject', 'subtree']

    def __init__(self, component, root_abspath,
        kbranch=None, prefix=None,
        is_root=False, is_toplevel=False, is_subproject=False,
        subproject=None, subtree=None):
        self.component = component
        self.kbranch = kbranch
        self.root_abspath = root_abspath
        self.prefix = prefix
        self.is_root = is_root
        self.is_toplevel = is_toplevel
        self.is_subproject = is_subproject
        self.subproject = subproject
        self.subtree = subtree

    name = property(lambda self: self.component.name)

    def _get_relpath(self):
        if self.prefix is not None:
            if self.component.relpath == '.':
                return self.prefix
            return osutils.pathjoin(self.prefix, self.component.relpath)
        else:
            return self.component.relpath

    relpath = property(_get_relpath)

    def _get_abspath(self):
        if self.component.relpath == '.':
            return self.root_abspath
        return osutils.pathjoin(self.root_abspath, self.component.relpath)

    abspath = property(_get_abspath)

    def __repr__(self):
        r = []
        if not self.is_subproject:
            r.append('Component(name=')
            r.append(self.name)
            r.append(',relpath=')
            r.append(self.relpath)
            r.append(',abspath=')
            r.append(self.abspath)
            if self.kbranch.url:
                r.append(',url=')
                r.append(self.kbranch.url)
            if self.kbranch.revision:
                r.append(',rev=')
                r.append(self.kbranch.revision)
            if self.is_root:
                r.append(',is_root')
            if self.is_toplevel:
                r.append(',is_toplevel')
            r.append(')')
        else:
            r.append('Subproject(name=')
            r.append(self.name)
            r.append(',relpath=')
            r.append(self.relpath)
            r.append(',abspath=')
            r.append(self.abspath)
            if self.component.url:
                r.append(',url=')
                r.append(self.component.url)
            r.append(',is_subproject')
            r.append(')')
        return ''.join(r)

    def _check(self, checklist):
        for ix, (name, path) in enumerate(checklist):
            if self.component.name == name or self.abspath == path:
                #checklist.pop(ix)
                return True
        return False

    def check(self, includechecklist, excludechecklist):
        if includechecklist:
            return self._check(includechecklist)
        if excludechecklist:
            return not self._check(excludechecklist)
        return True


class ProjectTree(object):
    """Represents all components and subprojects as a hierarchical tree."""

    __slots__ = ['project', 'alt_dict', 'toplevel', 'root_abspath', 'tree',
        'root_co', 'root_kb', 'prefix', 'root_name', 'root_revision']

    def __init__(self, proj, toplevel=False, prefix=None, root_name=None,
        root_revision=None):
        self.project = proj
        self.alt_dict = self.project.cfg.get_alt(self.project.get_alt_name())
        self.toplevel = toplevel
        self.prefix = prefix
        self.root_name = root_name
        self.root_revision = root_revision
        self.root_abspath = urlutils.local_path_from_url(self.project.t.base)
        if self.root_abspath.count('/') > 1 and self.root_abspath.endswith('/'):
            self.root_abspath = self.root_abspath[:-1]
        self.tree = []

    def build_tree(self, recurse=True):
        # create root component first
        self.root_co = RootComponent(self.project, self.root_name,
            revision=self.root_revision)
        self.root_kb = self.root_co.get_branch()
        self.tree = [ProjectNode(self.root_co, self.root_abspath,
            kbranch=self.root_kb, prefix=self.prefix,
            is_root=True, is_toplevel=self.toplevel)]
        # add components
        seen = {}
        for kname in self.project.get_component_list():
            ko = Component(self.project.cfg, kname, self.project.index_card)
            self._check_relpath(ko, seen)
            bname = self.alt_dict[kname]
            kb = ko.get_branch(bname)
            self.tree.append(ProjectNode(ko, self.root_abspath, kbranch=kb,
                prefix=self.prefix))
        # add subprojects
        for sname in self.project.get_subprojects_list():
            so = Subproject(self.project.cfg, sname, self.root_abspath,
                root_revision_id=self.project.index_card.get_revid(sname))
            self._check_relpath(so, seen)
            subp = so.get_instance()
            subtree = None
            if subp is not None and recurse:
                prefix = so.relpath
                if self.prefix is not None:
                    prefix = osutils.pathjoin(self.prefix, prefix)
                subtree_revision = self.project.index_card.get_revision(sname)
                subtree = ProjectTree(subp, prefix=prefix, root_name=sname,
                    root_revision=subtree_revision)
                subtree.build_tree()
            self.tree.append(ProjectNode(so, self.root_abspath,
                prefix=self.prefix,
                is_subproject=True, subproject=subp, subtree=subtree))

    def _check_relpath(self, component, seen):
        relpath = component.relpath
        if self.project.format == 'v2' and relpath == '.':
            raise errors.BzrError(
                "Project v2 can't have exlicit component at the root of "
                "the project (check component `%s`)" % component.name)
        if relpath in seen:
            raise errors.BzrError("Two components have the same relpath. "
                "Check components `%s` and `%s`" % (seen[relpath], component.name))
        seen[relpath] = component.name

    def iterate_nodes(self, yield_subproject_node=False, recurse=True):
        for node in self.tree:
            if not node.is_subproject:
                yield node
            else:
                if yield_subproject_node:
                    yield node
                if recurse and node.subtree is not None:
                    for subnode in node.subtree.iterate_nodes(
                                            yield_subproject_node, recurse):
                        yield subnode

    def iterate_nodes_reverse(self, yield_subproject_node=False, recurse=True):
        nodes = list(self.iterate_nodes(yield_subproject_node, recurse))
        for node in reversed(nodes):
            yield node

    def iterate_filter(self, yield_subproject_node=False, recurse=True,
        includes=None, excludes=None):
        if not includes and not excludes:
            return self.iterate_nodes(yield_subproject_node, recurse)
        if includes:
            includechecklist = ProjectTree.build_checklist(includes)
            excludechecklist = []
        else:
            includechecklist = []
            excludechecklist = ProjectTree.build_checklist(excludes)
        return self.iterate_filter_chk(yield_subproject_node, recurse,
                                       includechecklist, excludechecklist)

    @staticmethod
    def build_checklist(alist):
        return [(i, osutils.abspath(i)) for i in (alist or [])]

    def iterate_filter_chk(self, yield_subproject_node=False, recurse=True,
        includechecklist=None, excludechecklist=None):
        for node in self.iterate_nodes(True, False):
            check = node.check(includechecklist, excludechecklist)
            if not node.is_subproject:
                if check:
                    yield node
            else:
                if check:
                    if yield_subproject_node:
                        yield node
                    if recurse and node.subtree is not None:
                        # if subproject checked than all its components are also checked
                        for subnode in node.subtree.iterate_nodes(
                                            yield_subproject_node, recurse):
                            yield subnode

    def iterate_filter_reverse(self, yield_subproject_node=False, recurse=True,
        includes=None, excludes=None):
        nodes = list(self.iterate_filter(yield_subproject_node, recurse,
                                         includes, excludes))
        for node in reversed(nodes):
            yield node

    def iterate_filter_chk_reverse(self, yield_subproject_node=False,
        recurse=True, includechecklist=None, excludechecklist=None):
        nodes = list(self.iterate_filter_chk(yield_subproject_node, recurse,
            includechecklist, excludechecklist))
        for node in reversed(nodes):
            yield node

    def collect_revids(self):
        revids = {}
        for node in self.iterate_nodes(yield_subproject_node=True,
                                       recurse=False):
           if not node.is_root:
               path = node.abspath
               if bzr_vcs.is_checkout(path):
                   revids[node.name] = bzr_vcs.get_tip_revid(path)
        return revids


class SnapshotV2(object):

    def __init__(self, root_path):
        self.root_path = root_path
        self.index_path = osutils.pathjoin(self.root_path, SNAPSHOT_V2)
        self._index_card = None

    def get_index_card(self):
        if osutils.isfile(self.index_path):
            f = open(self.index_path, 'rU')
            try:
                self._index_card = config.IndexCard(f.read())
            finally:
                f.close()
        else:
            self._index_card = config.IndexCard()
        return self._index_card

    def set_index_card(self, card):
        f = open(self.index_path, 'wb')
        try:
            f.write(card.to_str())
            self._index_card = card
        finally:
            f.close()
        bzr_vcs.add(self.root_path, SNAPSHOT_V2)

    def update_index_card(self, revids=None):
        card = self.get_index_card()
        if card.update_revids(revids):
            self.set_index_card(card)
            return True
        return False

#-----------------------------------------------------------------------------
# uninteresting internal stuff

_WINDOWS_RESERVED_FILENAMES = set('''con prn aux nul
    com1 com2 com3 com4 com5 com6 com7 com8 com9
    lpt1 lpt2 lpt3 lpt4 lpt5 lpt6 lpt7 lpt8 lpt9'''.split())

_SAFE_MAP = {}

def _create_safe_map():
    global _SAFE_MAP
    _SAFE_MAP.update(dict((i, u'~%02x' % i) for i in (
        ord(s) for s in ' \\:*?"<>|/~')))                       # reserved characters
    _SAFE_MAP.update(dict((i, u'_'+s.lower()) for i, s in
        ((i, chr(i)) for i in range(ord('A'), ord('Z')+1))))    # lower the caps lock
    _SAFE_MAP.update({ord('_'): u'__'})                         # special case of caps lock

_create_safe_map()
