# Simple support for organizing bzr branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

from cStringIO import StringIO
import os

from bzrlib import (
    errors,
    trace,
    )
from bzrlib.tests import TestCase, TestCaseWithTransport

from bzrlib.plugins.scmproj.project import (
    Component,
    )
from bzrlib.plugins.scmproj.config import (
    BadComponentName,
    CONFIG_TMPL,
    Config,
    ConfigObjEx,
    LocalSettings,
    TemplateEngine,
    )
from bzrlib.plugins.scmproj.tests.whiteutils import reindentBlock


class TestConfigObjEx(TestCase):

    def test_space_in_section_name(self):
        s = reindentBlock("""\
            ### project.cfg v1

            [PROJECT]
            name = woo-hoo!
            
            [component "foo"]
            RELPATH = .
            
            [component "bar"]
            RELPATH = lib
            
            [subproject "utilities"]
            RELPATH = tools
            """)
        cfg = ConfigObjEx.from_str(s)
        self.assertEqual(s, cfg.to_str())


class TestLocalSettings(TestCaseWithTransport):

    def test_init_empty(self):
        wt = self.make_branch_and_tree('.')
        cfg = LocalSettings(wt.branch)
        self.assertEqual(None, cfg.project_url)
        self.assertEqual(None, cfg.alt)
        self.assertEqual(None, cfg.subset)
        self.assertEqual(None, cfg.workspace)

    def assertContains(self, haystack, needle):
        self.assertNotEqual(str(haystack).find(needle), -1,
            "Could not find `%s` in `%s`." % (needle, haystack))

    def test_modify(self):
        wt = self.make_branch_and_tree('.')
        cfg = LocalSettings(wt.branch)
        cfg.project_url = 'http://example.com/project'
        cfg.alt = 'trunk'
        cfg.subset = 'full'
        self.assertEqual('http://example.com/project', cfg.project_url)
        self.assertEqual('trunk', cfg.alt)
        self.assertEqual('full', cfg.subset)


class TestConfig(TestCase):

    def test_init_empty(self):
        cfg = Config()
        self.assertEqual(None, cfg.name)
        self.assertEqual(None, cfg.description)
        self.assertEqual([], cfg.components)
        self.assertEqual(None, cfg.vcs)

    def test_to_str_empty(self):
        cfg = Config()
        s = reindentBlock("""\
            ### project.cfg v1
            ### Don't edit the line above!

            """) + CONFIG_TMPL
        self.assertEqual(s, cfg.to_str())

    def test_init_default(self):
        cfg = Config(CONFIG_TMPL)
        self.assertEqual(None, cfg.name)
        self.assertEqual(None, cfg.description)
        self.assertEqual([], cfg.components)
        self.assertEqual('bzr', cfg.vcs)

    def test_from_str(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [PROJECT]
            [ALTS]
            [SUBSETS]
            """))
        self.assertEqual(None, cfg.name)
        self.assertEqual(None, cfg.description)
        self.assertEqual([], cfg.components)
        self.assertEqual('bzr', cfg.vcs)
        #
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [PROJECT]
            NAME =
            DESCRIPTION =
            VCS =
            
            [ALTS]
            
            [SUBSETS]
            """))
        self.assertEqual('', cfg.name)
        self.assertEqual('', cfg.description)
        self.assertEqual([], cfg.components)
        self.assertEqual('', cfg.vcs)
        #
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [PROJECT]
            NAME = test
            DESCRIPTION = Some descriptive description
            VCS = xyz
            
            [ALTS]
            
            [SUBSETS]

            [component "foo"]

            [component "bar"]
            """))
        self.assertEqual('test', cfg.name)
        self.assertEqual('Some descriptive description', cfg.description)
        self.assertEqual(['foo', 'bar'], cfg.components)
        self.assertEqual('xyz', cfg.vcs)

    def test_check_version_marker(self):
        cfg = Config()
        sio = StringIO()
        log_memento = trace.push_log_file(sio, log_format="%(message)s")
        try:
            cfg.from_str('')
        finally:
            trace.pop_log_file(log_memento)
        self.assertEqual('', sio.getvalue())

        cfg = Config()
        sio = StringIO()
        log_memento = trace.push_log_file(sio, log_format="%(message)s")
        try:
            cfg.from_str(reindentBlock("""\
                ### project.cfg v1

                [component "foo"]
                """))
        finally:
            trace.pop_log_file(log_memento)
        self.assertEqual('', sio.getvalue())

        cfg = Config()
        sio = StringIO()
        log_memento = trace.push_log_file(sio, log_format="%(message)s")
        try:
            cfg.from_str(reindentBlock("""\
                ###

                [component "foo"]
                """))
        finally:
            trace.pop_log_file(log_memento)
        self.assertContainsRe(sio.getvalue(),
            'WARNING: First line of project.cfg has no version marker')

    def test_get_subset_default_link(self):
        cfg = Config('')
        self.assertEqual('DEFAULT', cfg.get_subset_default_link())
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [SUBSETS]
            spam = foo
            eggs = bar
            xxx = bar foo
            zzz =
            """))
        self.assertEqual('DEFAULT', cfg.get_subset_default_link())
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [SUBSETS]
            DEFAULT = foo
            spam = foo
            eggs = bar
            xxx = bar foo
            zzz =
            """))
        self.assertEqual('DEFAULT', cfg.get_subset_default_link())
        # DEFAULT is the "link" to another subset
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [SUBSETS]
            DEFAULT = @spam
            spam = foo
            eggs = bar
            xxx = bar foo
            zzz =
            """))
        self.assertEqual('spam', cfg.get_subset_default_link())

    def test_get_subset(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [SUBSETS]
            spam = foo
            eggs = bar
            xxx = bar foo
            zzz =
            
            [component "foo"]
            RELPATH = aaa
            
            [component "bar"]
            RELPATH = bbb
            
            [component "baz"]
            RELPATH = bbb/ccc
            """))
        self.assertEqual(['foo', 'bar', 'baz'], cfg.get_subset('DEFAULT'))
        self.assertEqual(['foo', 'bar', 'baz'], cfg.get_subset('FULL'))
        self.assertEqual([], cfg.get_subset('NULL'))
        self.assertEqual(['foo'], cfg.get_subset('spam'))
        self.assertEqual(['bar'], cfg.get_subset('eggs'))
        self.assertEqual(['foo', 'bar'], cfg.get_subset('xxx'))
        self.assertEqual([], cfg.get_subset('zzz'))
        # DEFAULT is specified explicitly
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [SUBSETS]
            DEFAULT = foo
            spam = foo
            eggs = bar
            xxx = bar foo
            zzz =
            
            [component "foo"]
            RELPATH = aaa
            
            [component "bar"]
            RELPATH = bbb
            """))
        self.assertEqual(['foo'], cfg.get_subset('DEFAULT'))
        self.assertEqual(['foo', 'bar'], cfg.get_subset('FULL'))
        self.assertEqual([], cfg.get_subset('NULL'))
        # DEFAULT is the "link" to another subset
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [SUBSETS]
            DEFAULT = @spam
            spam = foo
            eggs = bar
            xxx = bar foo
            zzz =
            
            [component "foo"]
            RELPATH = aaa
            
            [component "bar"]
            RELPATH = bbb
            """))
        self.assertEqual(['foo'], cfg.get_subset('DEFAULT'))

    def test_get_alt_default_link(self):
        cfg = Config('')
        self.assertEqual('DEFAULT', cfg.get_alt_default_link())
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [ALTS]
            spam = foo:br-foo bar:br-bar
            eggs = foo:trunk-foo bar:br-bar
            fish = foo:br-foo
            meat = bar:trunk-bar
            """))
        self.assertEqual('DEFAULT', cfg.get_alt_default_link())
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [ALTS]
            DEFAULT = foo:br-foo bar:br-bar
            eggs = foo:trunk-foo bar:br-bar
            """))
        self.assertEqual('DEFAULT', cfg.get_alt_default_link())
        # DEFAULT is a link to named alt
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [ALTS]
            DEFAULT = @eggs
            spam = foo:br-foo bar:br-bar
            eggs = foo:trunk-foo bar:br-bar
            fish = foo:br-foo
            meat = bar:trunk-bar
            """))
        self.assertEqual('eggs', cfg.get_alt_default_link())

    def test_get_alt(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [ALTS]
            spam = foo:br-foo bar:br-bar
            eggs = foo:trunk-foo bar:br-bar
            fish = foo:br-foo
            meat = bar:trunk-bar

            [component "foo"]

            [component "bar"]
            """))
        self.assertEqual({'foo':'DEFAULT', 'bar':'DEFAULT'},
            cfg.get_alt('DEFAULT'))
        self.assertEqual({'foo':'br-foo', 'bar':'br-bar'},
            cfg.get_alt('spam'))
        self.assertEqual({'foo':'trunk-foo', 'bar':'br-bar'},
            cfg.get_alt('eggs'))
        self.assertEqual({'foo':'br-foo', 'bar':'DEFAULT'},
            cfg.get_alt('fish'))
        self.assertEqual({'foo':'DEFAULT', 'bar':'trunk-bar'},
            cfg.get_alt('meat'))
        self.assertRaises(errors.BzrError, cfg.get_alt, 'invalid-alt')
        # explicit DEFAULT
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [ALTS]
            DEFAULT = foo:br-foo bar:br-bar
            eggs = foo:trunk-foo bar:br-bar

            [component "foo"]

            [component "bar"]
            """))
        self.assertEqual({'foo':'br-foo', 'bar':'br-bar'},
            cfg.get_alt('DEFAULT'))
        # DEFAULT is a link to named alt
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [ALTS]
            DEFAULT = @eggs
            spam = foo:br-foo bar:br-bar
            eggs = foo:trunk-foo bar:br-bar
            fish = foo:br-foo
            meat = bar:trunk-bar

            [component "foo"]

            [component "bar"]
            """))
        self.assertEqual({'foo':'trunk-foo', 'bar':'br-bar'},
            cfg.get_alt('DEFAULT'))

    def test_get_alt_template(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [ALTS]
            DEFAULT = {COMPONENT}:template-branch
            template_alt = {COMPONENT}:template-branch bar:override-template
            
            [component "foo"]
            RELPATH = .
            BRANCHES = template-branch override-template
            
            [component "bar"]
            RELPATH = .
            BRANCHES = template-branch override-template
            """))
        self.assertEqual({'foo':'template-branch', 'bar':'template-branch'},
            cfg.get_alt('DEFAULT'))
        self.assertEqual({'foo':'template-branch', 'bar':'override-template'},
            cfg.get_alt('template_alt'))

    def test_component_list(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [PROJECT]
            COMPONENTS = foo bar

            [COMPONENTS]
            RELPATH = {COMPONENT}
            BRANCH_URL = http://example.com/{COMPONENT}
            """))
        self.assertEqual(['foo', 'bar'], cfg.components)
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [PROJECT]
            COMPONENTS = foo bar

            [component "foo"]
            """))
        self.assertEqual(['foo', 'bar'], cfg.components)
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [PROJECT]
            COMPONENTS = foo bar

            [component "foo"]

            [component "bar"]
            """))
        self.assertEqual(['foo', 'bar'], cfg.components)
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [PROJECT]
            COMPONENTS = foo bar

            [component "foo"]

            [component "bar"]

            [component "spam"]
            """))
        self.assertRaises(errors.BzrError, cfg._get_component_list)

    def test_get_component_relpath_template(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [COMPONENTS]
            RELPATH = x/{COMPONENT}
            
            [component "foo"]
            RELPATH = override-relpath
            BRANCHES =
            
            [component "bar"]
            BRANCHES =
            
            [component "baz"]
            RELPATH = {COMPONENT}/y
            BRANCHES =
            """))
        self.assertEqual('override-relpath',
            Component(cfg, 'foo').relpath)
        self.assertEqual('x/bar',
            Component(cfg, 'bar').relpath)
        self.assertEqual('baz/y',
            Component(cfg, 'baz').relpath)

        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
            BRANCHES =
            """))
        self.assertRaises(errors.BzrError, Component(cfg, 'foo')._get_relpath)

    def test_get_component_branches(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [COMPONENTS]
            BRANCHES = trunk 3-qa 4-prod
            
            [component "foo"]
            BRANCHES = override-branches
            
            [component "bar"]
            """))
        self.assertEqual(['override-branches'],
            Component(cfg, 'foo').branches)
        self.assertEqual(['trunk','3-qa', '4-prod'],
            Component(cfg, 'bar').branches)

    def test_get_component_branch_url_template(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [COMPONENTS]
            RELPATH = rel_{COMPONENT}_path
            BRANCH_URL = server/{BRANCH}/{RELPATH}/{COMPONENT}
            
            [component "boo"]
            BRANCHES = trunk
            
            [component "foo"]
            BRANCHES = trunk
            
                [[branch "trunk"]]
                BRANCH_URL = override-url
            
            [component "bar"]
            BRANCHES = trunk
            BRANCH_URL = http://server/project/branch
            
            [component "baz"]
            BRANCHES = trunk
            
                [[branch "trunk"]]
                BRANCH_URL = override-url/{BRANCH}/{RELPATH}/{COMPONENT}
            """))
        self.assertEqual('override-url',
            Component(cfg, 'foo').get_branch().url)
        self.assertEqual('server/trunk/rel_boo_path/boo',
            Component(cfg, 'boo').get_branch().url)
        self.assertEqual('http://server/project/branch',
            Component(cfg, 'bar').get_branch().url)
        self.assertEqual('override-url/trunk/rel_baz_path/baz',
            Component(cfg, 'baz').get_branch().url)
        #todo: non-default and alt

    def test__collect_components(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "boo"]
            
            [ component "foo" ]
            
            [    component   "bar"]
            
            [component\t"baz"]

            [component boo_nq]

            [ component foo_nq ]

            [    component   bar_nq]

            [component\tbaz_nq]
            """))
        # XXX ConfigObj seems to strip leading and trailing spaces from section names.
        d = cfg._collect_components()
        self.assertEqual({
            'boo': 'component "boo"',
            'foo': 'component "foo"',
            'bar': 'component   "bar"',
            'baz': 'component\t"baz"',
            'boo_nq': 'component boo_nq',
            'foo_nq': 'component foo_nq',
            'bar_nq': 'component   bar_nq',
            'baz_nq': 'component\tbaz_nq',
            }, dict(d))
        self.assertEquals(['boo', 'foo', 'bar', 'baz',
            'boo_nq', 'foo_nq', 'bar_nq', 'baz_nq'], cfg.components)
        # unmatched quotes should produce error
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "boo]
            """))
        self.assertRaises(errors.BzrError, cfg._collect_components)
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component boo"]
            """))
        self.assertRaises(errors.BzrError, cfg._collect_components)

    def test__collect_branches(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
                [[branch "trunk"]]
                [[   branch   "qa"   ]]
                [[branch   "dev" ]]
                [[branch trunk_nq]]
                [[   branch   qa_nq   ]]
                [[branch   dev_nq ]]
            """))
        d = cfg._collect_branches('foo')
        self.assertEqual({
            'trunk': 'branch "trunk"',
            'qa': 'branch   "qa"',
            'dev': 'branch   "dev"',
            'trunk_nq': 'branch trunk_nq',
            'qa_nq': 'branch   qa_nq',
            'dev_nq': 'branch   dev_nq',
            }, dict(d))
        self.assertEquals(['trunk', 'qa', 'dev',
            'trunk_nq', 'qa_nq', 'dev_nq'], d.keys())
        # check unmatched quotes
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
                [[branch "trunk]]
            """))
        self.assertRaises(errors.BzrError, cfg._collect_branches, 'foo')
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
                [[branch trunk"]]
            """))
        self.assertRaises(errors.BzrError, cfg._collect_branches, 'foo')

    def test_get_component_branches(self):
        # explicit list
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
            BRANCHES = trunk qa dev
            """))
        self.assertEqual(['trunk', 'qa', 'dev'],
            cfg.get_component_branches('foo'))
        # implicit list
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
                [[branch "trunk"]]
                [[branch "qa"]]
                [[branch "dev"]]
            """))
        self.assertEqual(['trunk', 'qa', 'dev'],
            cfg.get_component_branches('foo'))
        # implicit list should subset of explicit
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
            BRANCHES = trunk qa dev
                [[branch "trunk"]]
                [[branch "qa"]]
                [[branch "dev"]]
            """))
        self.assertEqual(['trunk', 'qa', 'dev'],
            cfg.get_component_branches('foo'))
        # otherwise error raised
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
            BRANCHES = trunk qa
                [[branch "trunk"]]
                [[branch "qa"]]
                [[branch "dev"]]
            """))
        self.assertRaises(errors.BzrError, cfg.get_component_branches, 'foo')
        # and we smart enough to derive branch name from url
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
            RELPATH = .
            BRANCH_URL = http://example.com/repo/trunk
            """))
        self.assertEqual(['trunk'], cfg.get_component_branches('foo'))

    def test_everything_is_optional(self):
        # empty config is good config
        cfg = Config('')
        self.assertEqual(None, cfg.name)
        self.assertEqual(None, cfg.description)
        self.assertEqual([], cfg.components)
        self.assertEqual('bzr', cfg.vcs)
        self.assertEqual({}, cfg.get_alt('DEFAULT'))
        self.assertEqual([], cfg.get_subset('DEFAULT'))
        # one component
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
            RELPATH = bar
            BRANCH_URL = http://example.com/spam
            """))
        self.assertEqual(None, cfg.name)
        self.assertEqual(None, cfg.description)
        self.assertEqual(['foo'], cfg.components)
        self.assertEqual('bzr', cfg.vcs)
        self.assertEqual({'foo': 'DEFAULT'}, cfg.get_alt('DEFAULT'))
        self.assertEqual(['foo'], cfg.get_subset('DEFAULT'))
        # one component + template for components
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [COMPONENTS]
            RELPATH = {COMPONENT}
            BRANCH_URL = http://example.com/{COMPONENT}

            [component "foo"]
            """))
        self.assertEqual(None, cfg.name)
        self.assertEqual(None, cfg.description)
        self.assertEqual(['foo'], cfg.components)
        self.assertEqual('bzr', cfg.vcs)
        self.assertEqual({'foo': 'DEFAULT'}, cfg.get_alt('DEFAULT'))
        self.assertEqual(['foo'], cfg.get_subset('DEFAULT'))

    def test_get_default_format(self):
        cfg = Config("")
        self.assertEqual(None, cfg.get_default_format())
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [COMPONENTS]
            FORMAT = rich-root-pack
            """))
        self.assertEqual('rich-root-pack', cfg.get_default_format())

    def test_get_default_vcs(self):
        cfg = Config("")
        self.assertEqual('bzr', cfg.get_default_vcs())
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [COMPONENTS]
            VCS = lokum
            """))
        self.assertEqual('lokum', cfg.get_default_vcs())

    def test__collect_subprojects(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [subproject "boo"]
            
            [ subproject   "foo" ]
            """))
        # XXX ConfigObj seems to strip leading and trailing spaces from section names.
        d = cfg._collect_subprojects()
        self.assertEqual({
            'boo': 'subproject "boo"',
            'foo': 'subproject   "foo"',
            }, dict(d))
        self.assertEquals(['boo', 'foo'], cfg.subprojects)

    def test_name_regex(self):
        # https://bugs.launchpad.net/bzr-scmproj/+bug/458937
        # bad component name should raise error
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "1foo"]
            RELPATH = foo
            BRANCH_URL = bzr://foo

            [subproject "bar"]
            RELPATH = bar
            SUBPROJECT_URL = bzr://bar
            """))
        self.assertRaises(BadComponentName, cfg._collect_components)
        # bad subproject name should raise error
        cfg = Config(reindentBlock("""\
            ### project.cfg v1
            [component "foo"]
            RELPATH = foo
            BRANCH_URL = bzr://foo

            [subproject "1bar"]
            RELPATH = bar
            SUBPROJECT_URL = bzr://bar
            """))
        self.assertRaises(BadComponentName, cfg._collect_subprojects)
